import os
import re
import json
import hashlib
import argparse

from smda.Disassembler import Disassembler
from smda.common.SmdaReport import SmdaReport

from mcrit.client.McritClient import McritClient
from mcrit.storage.FamilyEntry import FamilyEntry
from mcrit.storage.SampleEntry import SampleEntry
from mcrit.storage.FunctionEntry import FunctionEntry


### Helper functionality for submissions

def is_pe_or_elf(input_path):
    """
    Don't perform a full validation, just check heuristically if the file starts with PE or ELF markings
    """
    data = b""
    with open(input_path, "rb") as fin:
        data = fin.read(4)
    if len(data) >= 4:
        is_elf = data[:4] == b"\x7FELF"
        is_pe = data[:2] == b"MZ"
        return is_elf or is_pe
    return False


def get_base_addr(input_path):
    """
    If provided path ends with 0x[0-9a-fA-F]{8,16}, we assume that this is an IMAGEBASE address. 
    """
    base_addr = None
    baddr_match = re.search(re.compile("0x(?P<base_addr>[0-9a-fA-F]{8,16})$"), input_path)
    if baddr_match:
        base_addr = int(baddr_match.group("base_addr"), 16)
    return base_addr


def sha256(content):
    return hashlib.sha256(content).hexdigest()


def readFileContent(file_path):
    file_content = b""
    with open(file_path, "rb") as fin:
        file_content = fin.read()
    return file_content


def getFamilyName(input_path):
    family_name = ""
    abs_path = os.path.abspath(input_path)
    for folder in abs_path.split("/")[::-1]:
        if folder:
            family_name = folder
    return family_name


def getSampleVersion(input_path, family):
    sample_version = ""
    collected = []
    for folder in input_path.split("/"):
        if folder != family and folder != "":
            collected.append(folder)
    if collected:
        sample_version = os.sep.join(collected)
    return sample_version


def getFolderFilePath(input_root, input_path):
    abs_path = os.path.abspath(input_path)
    relative_filepath = abs_path[len(input_root):]
    return relative_filepath


def getSmdaReportFromFilepath(args, filepath):
    filename = os.path.basename(filepath)
    smda_report = None
    if args.smda:
        if not filename.endswith(".smda"):
            print(f"Skipping a file not recognized as SMDA report: {filepath}")
        else:
            try:
                smda_report = SmdaReport.fromFile(filepath)
            except:
                print(f"Failed to parse SMDA report: {filepath}")
    else:
        if args.executables_only and not is_pe_or_elf(filepath):
            print(f"Skipping a file not recognized as executable: {filepath}")
        else:
            disassembler = Disassembler()
            if get_base_addr(filename) is not None:
                base_addr = get_base_addr(filename)
                try:
                    smda_report = disassembler.disassembleBuffer(readFileContent(filepath), base_addr)
                    smda_report.filename = filename
                except Exception as exc:
                    import traceback
                    print(f"ERROR: SMDA caused an exception while processing this file: {filepath}")
                    print(traceback.format_exc())
                    return None
            else:
                try:
                    smda_report = disassembler.disassembleFile(filepath)
                except:
                    import traceback
                    print(f"ERROR: SMDA caused an exception while processing this file: {filepath}")
                    print(traceback.format_exc())
                    return None
    # apply any of the forced flags: family, version, library
    if smda_report:
        if args.mode in ["file", "dir"]:
            if args.family is not None:
                smda_report.family = args.family
            if args.version is not None:
                smda_report.version = args.version
        if args.mode in ["file", "dir", "recursive"] and args.library:
            smda_report.is_library = True
        if args.output:
            with open() as f_smda:
                json.dump(smda_report.toDict(), f_smda, sort_keys=True, indent=1)
    return smda_report



### Main Processing 

class McritConsole(object):

    def __init__(self) -> None:
        self.parser = argparse.ArgumentParser(description="Console interaction with MCRIT.")
        subparsers = self.parser.add_subparsers(dest="command")
        self.parser_client = subparsers.add_parser("client")
        subparser_client = self.parser_client.add_subparsers(dest="client_command")
        client_status = subparser_client.add_parser("status")
        # client submit
        client_submit = subparser_client.add_parser("submit", help="Various ways of file submission incl. disassembly using SMDA if needed.")
        client_submit.add_argument("filepath", type=str, help="Submit the folllowing <filepath>, indicating a (file/dir).")
        client_submit.add_argument("--mode", type=str, default="file", choices=["file", "dir", "recursive", "malpedia"], help="Submit a single <file> or all files in a <dir>. Use <recursive> submission for a folder structured as ./family_name/version/version/files. Synchronize <malpedia> into MCRIT. Default: <file>.")
        client_submit.add_argument("-f", "--family", type=str, help="Set/Override SmdaReport with this family (only in modes: file/dir)")
        client_submit.add_argument("-v", "--version", type=str, help="Set/Override SmdaReport with this version (only in modes: file/dir)")
        client_submit.add_argument("-l", "--library", action="store_true", help="Set/Override SmdaReport with the library flag (only in modes: file/dir/recursive, default: False).")
        client_submit.add_argument("-x", "--executables_only", action="store_true", help="Only process files that are parsable PE or ELF files (default: False).")
        client_submit.add_argument("-o", "--output", type=str, help="Optionally store SMDA reports in folder OUTPUT.")
        client_submit.add_argument("-s", "--smda", action="store_true", help="Do not disassemble, instead only submit files that are recognized as SMDA reports (only works with modes: file/dir).")
        # client import 
        client_import = subparser_client.add_parser("import", help="Import of previously exported data in the MCRIT format.")
        client_import.add_argument("filepath", type=str, help="Import a given <filepath> containing MCRIT data into the storage.")
        client_export = subparser_client.add_parser("export", help="Export of data from MCRIT.")
        client_export.add_argument("filepath", type=str, help="Export the full storage into <filepath>.")
        client_export.add_argument("--sample_ids", type=str, help="Limit export to a list of comma-separated <sample_ids>.")
        client_search = subparser_client.add_parser("search", help="Search for arbitrary data based on a search string.")
        client_search.add_argument("search_term", type=str, help="Conduct a search with a given <search_term>.")
        client_queue = subparser_client.add_parser("queue", help="Inspection of the current processing queue.")
        client_queue.add_argument("--filter", type=str, help="Filter queue entries with this term.")

    def run(self):
        ARGS = self.parser.parse_args()
        if ARGS.client_command == "status":
            self._handle_status(ARGS)
        elif ARGS.client_command == "import":
            self._handle_import(ARGS)
        elif ARGS.client_command == "export":
            self._handle_export(ARGS)
        elif ARGS.client_command == "search":
            self._handle_search(ARGS)
        elif ARGS.client_command == "queue":
            self._handle_queue(ARGS)
        elif ARGS.client_command == "submit":
            self._handle_submit(ARGS)
        elif ARGS.client_command == "sync":
            self._handle_sync(ARGS)
        else:
            print(self.parser_client.print_help())

    def _handle_status(self, args):
        client = McritClient()
        result = client.getStatus()
        print(result)

    def _handle_import(self, args):
        client = McritClient()
        if os.path.isfile(args.filepath):
            with open(args.filepath) as fin:
                result = client.addImportData(json.load(fin))
                print(result)

    def _handle_export(self, args):
        client = McritClient()
        sample_ids = None
        if args.sample_ids is not None:
            sample_ids = [int(i) for i in args.sample_ids.split(",")]
        result = client.getExportData(sample_ids=sample_ids)
        with open(args.filepath, "w") as fout:
            json.dump(result, fout, indent=1)
        print(f"wrote export to {args.filepath}.")

    def _handle_search(self, args):
        client = McritClient()
        result = client.search_families(args.search_term)
        if result["search_results"]:
            print("Family Search Results")
            for family_id, entry in result["search_results"].items():
                family_entry = FamilyEntry.fromDict(entry)
                print(f"{family_entry}")
            print("*" * 20)
        result = client.search_samples(args.search_term)
        if result["search_results"]:
            print("Sample Search Results")
            for sample_id, entry in result["search_results"].items():
                sample_entry = SampleEntry.fromDict(entry)
                print(f"{sample_entry}")
            print("*" * 20)
        result = client.search_functions(args.search_term)
        if result["search_results"]:
            print("Function Search Results")
            for function_id, entry in result["search_results"].items():
                function_entry = FunctionEntry.fromDict(entry)
                print(f"{function_entry}")

    def _handle_queue(self, args):
        client = McritClient()
        result = client.getQueueData(filter=args.filter)
        for entry in result:
            job_id = entry.job_id
            result_id = entry.result
            created = entry.created_at if entry.created_at is not None else "-"
            started = entry.started_at if entry.started_at is not None else "-"
            finished = entry.finished_at if entry.finished_at is not None else "-"
            method = entry.parameters
            progress = entry.progress
            print(f"{job_id} {result_id} | {created} {started} {finished} | {method} - {progress}")

    def _handle_submit(self, args):
        # run a number of sanity checks first
        if not os.path.exists(args.filepath):
            print("Your <filepath> does not exist.")
            return
        if args.mode == "file" and not os.path.isfile(args.filepath):
            print("Mode <file> only works when <filepath> is a file.")
            return
        if args.mode in ["dir", "recursive", "malpedia"] and not os.path.isdir(args.filepath):
            print("Modes <dir|recursive|malpedia> only work when <filepath> is a directory.")
            return
        if args.output is not None and (not os.path.exists(args.output) or not os.path.isdir(args.output)):
            print("Your <output> is not a directory or does not exist.")
            return
        if args.smda and args.mode in ["recursive", "malpedia"]:
            print("Modes <recursive|malpedia> are not compatible with SMDA report loading.")
            return
        # behavior according to the modes offered
        if args.mode == "file":
            self._handle_submit_file(args)
        if args.mode == "dir":
            self._handle_submit_dir(args)
        if args.mode == "recursive":
            self._handle_submit_recursive(args)
        if args.mode == "malpedia":
            self._handle_submit_malpedia(args)

    def _handle_submit_file(self, args):
        client = McritClient()
        mcrit_samples = client.getSamples()
        mcrit_samples_by_sha256 = {}
        for sample_id, sample in mcrit_samples.items():
            mcrit_samples_by_sha256[sample.sha256] = sample
        if sha256(readFileContent(args.filepath)) in mcrit_samples_by_sha256:
            print(f"SKIPPING: {args.filepath} - already in MCRIT.")
        smda_report = getSmdaReportFromFilepath(args, args.filepath)
        if smda_report:
            print(smda_report)
            client.addReport(smda_report)

    def _handle_submit_dir(self, args):
        client = McritClient()
        mcrit_samples = client.getSamples()
        mcrit_samples_by_sha256 = {}
        for sample_id, sample in mcrit_samples.items():
            mcrit_samples_by_sha256[sample.sha256] = sample
        for filename in os.listdir(args.filepath):
            filepath = os.sep.join([args.filepath, filename])
            if os.path.isfile(filepath):
                if sha256(readFileContent(filepath)) in mcrit_samples_by_sha256:
                    print(f"SKIPPING: {filepath} - already in MCRIT.")
                    continue
                smda_report = getSmdaReportFromFilepath(args, filepath)
                if smda_report:
                    print(smda_report)
                    client.addReport(smda_report)

    def _handle_submit_recursive(self, args):
        client = McritClient()
        mcrit_samples = client.getSamples()
        mcrit_samples_by_sha256 = {}
        for sample_id, sample in mcrit_samples.items():
            mcrit_samples_by_sha256[sample.sha256] = sample
        for root, subdir, files in sorted(os.walk(args.filepath)):
            folder_relative_path = getFolderFilePath(args.filepath, root)
            for filename in files:
                filepath = os.sep.join([root, filename])
                if os.path.isfile(filepath):
                    if sha256(readFileContent(filepath)) in mcrit_samples_by_sha256:
                        print(f"SKIPPING: {filepath} - already in MCRIT.")
                        continue
                    smda_report = getSmdaReportFromFilepath(args, filepath)
                    if smda_report:
                        smda_report.family = getFamilyName(folder_relative_path)
                        smda_report.version = getSampleVersion(folder_relative_path, smda_report.family)
                        print(filepath)
                        print(smda_report)
                        client.addReport(smda_report)

    def _handle_submit_malpedia(self, args):
        # verify that we have a malpedia root
        malpedia_root = os.path.abspath(args.filepath)
        malpedia_root = malpedia_root.rstrip("/")
        if not malpedia_root.endswith("malpedia"):
            print(f"Error: You pointing to a folder named differently than 'malpedia'.")
            return
        if not "malpedia.bib" in os.listdir(malpedia_root):
            print(f"Error: 'malpedia.bib' is missing in that folder, are you sure you are poniting to a Malpedia repository?")
            return
        client = McritClient()
        # get current status of all samples in MCRIT
        mcrit_samples = client.getSamples()
        mcrit_samples_by_filename = {}
        for sample_id, sample in mcrit_samples.items():
            # verify that filename has malpedia format (starts with sha256 and _unpacked/_dump)
            if sample.filename in mcrit_samples_by_filename:
                print(f"WARNING: filename {sample.filename} appears to exist more than once in your MCRIT instance, now using SHA256 {sample.sha256[:8]}.")
            mcrit_samples_by_filename[sample.filename] = sample
        # get all unpacked/dumped files and their family/version by crawling given Malpedia location
        malpedia_samples_by_filename = self._getMalpediaSamplesByFilename(malpedia_root)
        # use submission filenames within MCRIT to check for presence of files and verify their family/version identity
        for filename, malpedia_info in malpedia_samples_by_filename.items():
            malpedia_family = malpedia_info["family"]
            malpedia_version = malpedia_info["version"]
            malpedia_filepath = malpedia_info["filepath"]
            if filename in mcrit_samples_by_filename:
                mcrit_family = mcrit_samples_by_filename[filename].family
                mcrit_version = mcrit_samples_by_filename[filename].version
                # warn about files were family/version mismatches.
                if (malpedia_family != mcrit_family or
                        malpedia_version != mcrit_version):
                    print(f"WARNING: Sample {mcrit_samples_by_filename[filename].sample_id} with filename {filename} has different family/version information: ")
                    print(f"* Malpedia: {malpedia_family}|{malpedia_version}")
                    print(f"* MCRIT: {mcrit_family}|{mcrit_version}")
            # directly add all files in Malpedia but missing in MCRIT
            else:
                smda_report = getSmdaReportFromFilepath(args, malpedia_filepath)
                if smda_report:
                    smda_report.family = malpedia_family
                    smda_report.version = malpedia_version
                    print(malpedia_filepath)
                    print(smda_report)
                    client.addReport(smda_report)
        # warn about files that appear deleted because not present in Malpedia but in MCRIT (based on name schema)
        for filename, mcrit_sample in mcrit_samples_by_filename.items():
            if self._isMalpediaFilename(filename) and filename not in malpedia_samples_by_filename:
                print(f"WARNING: Sample {mcrit_sample.sample_id} with filename {filename} ({mcrit_sample.family}|{mcrit_sample.version}) present in MCRIT but not in Malpedia?")

    def _getMalpediaSamplesByFilename(self, malpedia_root):
        malpedia_samples_by_filename = {}
        for root, subdir, files in sorted(os.walk(malpedia_root)):
            if not "win." in root or "elf." in root:
                continue
            folder_relative_path = getFolderFilePath(malpedia_root, root)
            for filename in sorted(files):
                if self._isMalpediaFilename(filename):
                    filepath = root + os.sep + filename
                    sample_family = getFamilyName(folder_relative_path)
                    malpedia_samples_by_filename[filename] = {
                        "filename": filename,
                        "filepath": filepath,
                        "family": sample_family,
                        "version": getSampleVersion(folder_relative_path, sample_family)
                    }
        return malpedia_samples_by_filename
    
    def _isMalpediaFilename(self, filename):
        malpedia_file_pattern = re.compile("^[0-9a-f]{64}(_unpacked|dump7?_0x[0-9a-fA-F]{8,16})")
        return re.search(malpedia_file_pattern, filename)



import functools
import time
import json
import logging
from typing import Dict, List, Optional, Tuple

import requests
import urllib.parse
from mcrit.storage.FamilyEntry import FamilyEntry
from mcrit.storage.FunctionEntry import FunctionEntry
from mcrit.storage.SampleEntry import SampleEntry
from mcrit.queue.LocalQueue import Job
from smda.common.SmdaReport import SmdaReport
from smda.Disassembler import Disassembler


# Only do basicConfig if no handlers have been configured
if len(logging._handlerList) == 0:
    logging.basicConfig(level=logging.INFO, format="%(asctime)-15s %(message)s")
LOGGER = logging.getLogger(__name__)


class JobTerminatedError(Exception):
    pass


def isJobTerminated(job):
    if job is None:
        return True

    return job.is_terminated

def isJobFailed(job):
    return (job is not None) and (not job.is_failed)

def isJobFinishedTerminatedOrFailed(job):
    return isJobTerminated(job) or (job.result is not None) or isJobFailed(job)


def handle_response(response):
    data = None
    if response.status_code in [500, 501]:
        LOGGER.warning("McritClient received status code 500 from MCRIT.")
    elif response.status_code in [400, 404, 410]:
        # nothing to here as of now
        pass
    elif response.status_code in [200, 202]:
        json_response = response.json()
        if "status" in json_response and json_response["status"] == "successful":
            data = json_response["data"]
    return data


class McritClient:
    def __init__(self, mcrit_server=None, apitoken=None, username=None, raw_responses=False):
        self.mcrit_server = "http://localhost:8000"
        self.headers = {}
        self.raw = True if raw_responses else False
        if apitoken:
            self.headers.update({"apitoken": apitoken})
        if username:
            self.headers.update({"username": username})
        if mcrit_server is not None:
            self.mcrit_server = mcrit_server

    def setApitoken(self, apitoken):
        self.headers.update({"apitoken": apitoken})

    def setUsername(self, username):
        self.headers.update({"username": username})

    def _getMatchingRequestParams(
        self, minhash_threshold=None, pichash_size=None, force_recalculation=None, band_matches_required=None, exclude_self_matches=False
    ):
        params = {}
        if minhash_threshold is not None:
            params["minhash_score"] = minhash_threshold
        if pichash_size is not None:
            params["pichash_size"] = pichash_size
        if force_recalculation is not None:
            params["force_recalculation"] = force_recalculation
        if band_matches_required is not None:
            params["band_matches_required"] = band_matches_required
        if exclude_self_matches:
            params["exclude_self_matches"] = True
        return params

    def respawn(self):
        response = requests.post(f"{self.mcrit_server}/respawn", headers=self.headers)
        return handle_response(response)
    
    def completeMinhashes(self):
        response = requests.get(f"{self.mcrit_server}/complete_minhashes", headers=self.headers)
        return handle_response(response)

    def rebuildIndex(self):
        response = requests.get(f"{self.mcrit_server}/rebuild_index", headers=self.headers)
        return handle_response(response)

    def addReport(self, smda_report: SmdaReport) -> Tuple[SampleEntry, Optional[str]]:
        smda_json = smda_report.toDict()
        response = requests.post(f"{self.mcrit_server}/samples", json=smda_json, headers=self.headers)
        if self.raw:
            return response
        data = handle_response(response)
        if data is not None:
            if "job_id" in data:
                job_id = data["job_id"]
            else:
                job_id = None
            return SampleEntry.fromDict(data["sample_info"]), job_id

    def addBinarySample(self, binary: bytes, filename=None, family=None, version=None, is_dump=False, base_addr=None, bitness=None) -> Tuple[SampleEntry, Optional[str]]:
        query_fields = []
        if filename is not None:
            query_fields.append(f"filename={filename}")
        if family is not None:
            query_fields.append(f"family={family}")
        if version is not None:
            query_fields.append(f"version={version}")
        if is_dump:
            query_fields.append(f"is_dump=1")
        if base_addr is not None:
            query_fields.append(f"base_addr=0x{base_addr:x}")
        if bitness is not None and bitness in [32, 64]:
            query_fields.append(f"bitness={bitness}")
        query_string = ""
        if len(query_fields) > 0:
            query_string = "?" + "&".join(query_fields)
        response = requests.post(f"{self.mcrit_server}/samples/binary{query_string}", data=binary, headers=self.headers)
        return handle_response(response)

    ###########################################
    ### Families 
    ###########################################

    def modifyFamily(self, family_id, family_name=None, is_library=None):
        update_dict = {}
        if family_name is not None:
            update_dict["family_name"] = family_name
        if is_library is not None:
            update_dict["is_library"] = is_library
        response = requests.put(f"{self.mcrit_server}/families/{family_id}", update_dict, headers=self.headers)
        return handle_response(response)

    def getFamily(self, family_id: int, with_samples=True) -> Optional[FamilyEntry]:
        """
        Get a FamilyEntry by its <family_id>
        Supported by mcritweb API pass-through
        """
        query_params = "?with_samples=true" if with_samples else "?with_samples=false"
        response = requests.get(f"{self.mcrit_server}/families/{family_id}{query_params}", headers=self.headers)
        if self.raw:
            return response
        data = handle_response(response)
        if data is not None:
            return FamilyEntry.fromDict(data)
        return None

    def getFamilies(self) -> Optional[Dict[int, FamilyEntry]]:
        """
        Get all FamilyEntry objects in a dict, with <family_id> as key
        Supported by mcritweb API pass-through
        """
        response = requests.get(f"{self.mcrit_server}/families", headers=self.headers)
        if self.raw:
            return response
        data = handle_response(response)
        if data is not None:
            return {i: FamilyEntry.fromDict(entry) for i, entry in data.items()}
        return None

    def isFamilyId(self, family_id) -> bool:
        """
        Check if a <family_id> is valid in MCRIT
        Supported by mcritweb API pass-through
        """
        response = requests.get(f"{self.mcrit_server}/families/{family_id}", headers=self.headers)
        if self.raw:
            return response
        data = handle_response(response)
        if data is not None:
            return True
        return False

    def deleteFamily(self, family_id, keep_samples=False):
        query_params = "?keep_samples=true" if keep_samples else "?keep_samples=false"
        response = requests.delete(f"{self.mcrit_server}/families/{family_id}{query_params}", headers=self.headers)
        return handle_response(response)

    ###########################################
    ### Samples 
    ###########################################

    def isSampleId(self, sample_id):
        """
        Check if a <sample_id> is valid in MCRIT
        Supported by mcritweb API pass-through
        """
        response = requests.get(f"{self.mcrit_server}/samples/{sample_id}", headers=self.headers)
        if self.raw:
            return response
        data = handle_response(response)
        if data is not None:
            return True
        return False

    def modifySample(self, sample_id, family_name=None, version=None, component=None, is_library=None):
        update_dict = {}
        if family_name is not None:
            update_dict["family_name"] = family_name
        if version is not None:
            update_dict["version"] = version
        if component is not None:
            update_dict["component"] = component
        if is_library is not None:
            update_dict["is_library"] = is_library
        response = requests.put(f"{self.mcrit_server}/samples/{sample_id}", update_dict, headers=self.headers)
        return handle_response(response)

    def deleteSample(self, sample_id):
        response = requests.delete(f"{self.mcrit_server}/samples/{sample_id}", headers=self.headers)
        return handle_response(response)

    def getSamplesByFamilyId(self, family_id: int) -> Optional[List[SampleEntry]]:
        family_data = self.getFamily(family_id)
        if family_data is not None:
            return family_data.samples

    def getSampleById(self, sample_id):
        """
        Get a SampleEntry by its <sample_id>
        Supported by mcritweb API pass-through
        """
        response = requests.get(f"{self.mcrit_server}/samples/{sample_id}", headers=self.headers)
        if self.raw:
            return response
        data = handle_response(response)
        if data is not None:
            return SampleEntry.fromDict(data)

    def getSamples(self, start=0, limit=0):
        """
        Get all SampleEntries, optionally from sample_id <start> and up to <limit> many
        Supported by mcritweb API pass-through
        """
        query_string = ""
        if (isinstance(start, int) and start >= 0) and (isinstance(limit, int) and limit >= 0):
            query_string = f"?start={start}&limit={limit}"
        response = requests.get(f"{self.mcrit_server}/samples{query_string}", headers=self.headers)
        if self.raw:
            return response
        data = handle_response(response)
        if data is not None:
            return {int(k): SampleEntry.fromDict(v) for k, v in data.items()}

    ###########################################
    ### Functions
    ###########################################

    def getFunctionsBySampleId(self, sample_id):
        """
        Get a all FunctionEntries for a given <sample_id>
        Supported by mcritweb API pass-through
        """
        response = requests.get(f"{self.mcrit_server}/samples/{sample_id}/functions", headers=self.headers)
        if self.raw:
            return response
        data = handle_response(response)
        if data is not None:
            return [
                FunctionEntry.fromDict(function_entry_dict)
                for function_entry_dict in data.values()
            ]

    def getFunctions(self, start=0, limit=0):
        """
        Get a all FunctionEntries, optionally from sample_id <start> and up to <limit> many
        Supported by mcritweb API pass-through
        """
        query_string = ""
        if (isinstance(start, int) and start >= 0) and (isinstance(limit, int) and limit >= 0):
            query_string = f"?start={start}&limit={limit}"
        response = requests.get(f"{self.mcrit_server}/functions{query_string}", headers=self.headers)
        if self.raw:
            return response
        data = handle_response(response)
        if data is not None:
            return {int(k): FunctionEntry.fromDict(v) for k, v in data.items()}
        
    def getFunctionsByIds(self, function_ids:list, with_label_only=False):
        """
        Get all FunctionEntries identified by the provided list of function_ids
        Supported by mcritweb API pass-through
        """
        query_with_label_only = "?with_label_only=True" if with_label_only else ""
        function_id_string = ",".join(["%d" % fid for fid in function_ids])
        response = requests.post(f"{self.mcrit_server}/functions{query_with_label_only}", data=function_id_string, headers=self.headers)
        if self.raw:
            return response
        data = handle_response(response)
        if data is not None:
            return {int(k): FunctionEntry.fromDict(v) for k, v in data.items()}
        return {}

    def isFunctionId(self, function_id):
        """
        Check if a <function_id> is valid in MCRIT
        Supported by mcritweb API pass-through
        """
        response = requests.get(f"{self.mcrit_server}/functions/{function_id}", headers=self.headers)
        if self.raw:
            return response
        data = handle_response(response)
        if self.raw:
            return data
        if data is not None:
            return True
        return False

    def getFunctionById(self, function_id: int, with_xcfg=False) -> Optional[FunctionEntry]:
        """
        Get a FunctionEntry by its <function_id>
        Supported by mcritweb API pass-through
        """
        query_with_xcfg = "?with_xcfg=True" if with_xcfg else ""
        response = requests.get(f"{self.mcrit_server}/functions/{function_id}{query_with_xcfg}", headers=self.headers)
        data = handle_response(response)
        if self.raw:
            return response
        if data is not None:
            return FunctionEntry.fromDict(data)

    def completeMinhashes(self):
        response = requests.get(f"{self.mcrit_server}/complete_minhashes", headers=self.headers)
        return handle_response(response)

    ###########################################
    ### Matching 
    ###########################################

    def requestMatchesForSmdaReport(
        self,
        smda_report: SmdaReport,
        minhash_threshold=None,
        pichash_size=None,
        band_matches_required=None,
        force_recalculation=False,
    ) -> str:
        smda_json = smda_report.toDict()
        params = self._getMatchingRequestParams(
            minhash_threshold, pichash_size, force_recalculation, band_matches_required
        )
        response = requests.post(f"{self.mcrit_server}/query", json=smda_json, headers=self.headers, params=params)
        return handle_response(response)

    def requestMatchesForMappedBinary(
        self,
        binary: bytes,
        base_address: int,
        minhash_threshold=None,
        pichash_size=None,
        band_matches_required=None,
        disassemble_locally=True,
        force_recalculation=False,
    ) -> str:
        if disassemble_locally:
            disassembler = Disassembler()
            smda_report = disassembler.disassembleBuffer(binary, base_address)
            if smda_report.status == "error":
                return None
            return self.requestMatchesForSmdaReport(
                smda_report,
                minhash_threshold=minhash_threshold,
                pichash_size=pichash_size,
                band_matches_required=band_matches_required,
                force_recalculation=force_recalculation,
            )

        params = self._getMatchingRequestParams(
            minhash_threshold, pichash_size, force_recalculation, band_matches_required
        )
        response = requests.post(f"{self.mcrit_server}/query/binary/mapped/{base_address}", binary, headers=self.headers, params=params)
        return handle_response(response)

    def requestMatchesForUnmappedBinary(
        self,
        binary: bytes,
        minhash_threshold=None,
        pichash_size=None,
        band_matches_required=None,
        disassemble_locally=True,
        force_recalculation=False,
    ) -> str:
        if disassemble_locally:
            disassembler = Disassembler()
            smda_report = disassembler.disassembleUnmappedBuffer(binary)
            if smda_report.status == "error":
                return None
            return self.requestMatchesForSmdaReport(
                smda_report,
                minhash_threshold=minhash_threshold,
                pichash_size=pichash_size,
                band_matches_required=band_matches_required,
                force_recalculation=force_recalculation,
            )

        params = self._getMatchingRequestParams(
            minhash_threshold, pichash_size, force_recalculation, band_matches_required
        )

        response = requests.post(f"{self.mcrit_server}/query/binary", binary, headers=self.headers, params=params)
        return handle_response(response)

    def requestMatchesForSample(
        self,
        sample_id,
        minhash_threshold=None,
        pichash_size=None,
        band_matches_required=None,
        force_recalculation=False,
    ) -> None:
        params = self._getMatchingRequestParams(
            minhash_threshold, pichash_size, force_recalculation, band_matches_required
        )
        response = requests.get(f"{self.mcrit_server}/matches/sample/{sample_id}", headers=self.headers, params=params)
        if self.raw:
            return response
        return handle_response(response)

    def requestMatchesForSampleVs(
        self,
        sample_id,
        other_sample_id,
        minhash_threshold=None,
        pichash_size=None,
        band_matches_required=None,
        force_recalculation=False,
    ) -> str:
        params = self._getMatchingRequestParams(
            minhash_threshold, pichash_size, force_recalculation, band_matches_required
        )
        response = requests.get(f"{self.mcrit_server}/matches/sample/{sample_id}/{other_sample_id}", headers=self.headers, params=params)
        if self.raw:
            return response
        return handle_response(response)

    def requestMatchesCross(
        self,
        sample_ids,
        minhash_threshold=None,
        pichash_size=None,
        band_matches_required=None,
        force_recalculation=False,
    ) -> None:
        params = self._getMatchingRequestParams(
            minhash_threshold, pichash_size, force_recalculation, band_matches_required
        )
        response = requests.get(
            f"{self.mcrit_server}/matches/sample/cross/{','.join([str(id) for id in sample_ids])}", 
            headers=self.headers,
            params=params
        )
        return handle_response(response)

    def getMatchFunctionVs(
        self,
        function_id_a:int,
        function_id_b:int
    ) -> None:
        response = requests.get(f"{self.mcrit_server}/matches/function/{function_id_a}/{function_id_b}", headers=self.headers)
        if self.raw:
            return response
        return handle_response(response)


    def getMatchesForSmdaFunction(self, smda_report, minhash_threshold=None, pichash_size=None, force_recalculation=None, band_matches_required=None, exclude_self_matches=False):
        """
        Get all matches for a SmdaReport with a single SmdaFunction
        Supported by mcritweb API pass-through
        """
        # TODO add the same parameter possibilities that are used for regular full matching jobs
        params = self._getMatchingRequestParams(
            minhash_threshold, pichash_size, force_recalculation, band_matches_required, exclude_self_matches
        )
        response = requests.post(f"{self.mcrit_server}/query/function", json=smda_report.toDict(), headers=self.headers, params=params)
        if self.raw:
            return response
        return handle_response(response)

    def getMatchesForPicHash(self, pichash, summary=False):
        """
        Get all matches for a given <pichash>, optionally only as <summary>
        Supported by mcritweb API pass-through
        """
        summary_string = "/summary" if summary else ""
        response = requests.get(f"{self.mcrit_server}/query/pichash/{pichash:016x}{summary_string}", headers=self.headers)
        if self.raw:
            return response
        return handle_response(response)

    def getMatchesForPicBlockHash(self, picblockhash, summary=False):
        """
        Get all matches for a given <picblockhash>, optionally only as <summary>
        Supported by mcritweb API pass-through
        """
        summary_string = "/summary" if summary else ""
        response = requests.get(f"{self.mcrit_server}/query/picblockhash/{picblockhash:016x}{summary_string}", headers=self.headers)
        if self.raw:
            return response
        return handle_response(response)

    def getSampleBySha256(self, sample_sha256: str):
        """
        Get a SampleEntry by its <sha256>
        Supported by mcritweb API pass-through
        """
        response = requests.get(f"{self.mcrit_server}/samples/sha256/{sample_sha256}", headers=self.headers)
        if self.raw:
            return response
        data = handle_response(response)
        if data is None:
            return None
        return SampleEntry.fromDict(data)

    ###########################################
    ### Status, Results 
    ###########################################

    def getStatus(self):
        """
        Get a status report of the MCRIT server with some statistics
        Supported by mcritweb API pass-through
        """
        response = requests.get(f"{self.mcrit_server}/status", headers=self.headers)
        if self.raw:
            return response
        return handle_response(response)

    def getVersion(self):
        """
        Get a version report of the MCRIT server
        Supported by mcritweb API pass-through
        """
        response = requests.get(f"{self.mcrit_server}/version", headers=self.headers)
        if self.raw:
            return response
        data = handle_response(response)
        if isinstance(data, dict) and "version" in data:
            return data["version"]
        return None

    def getJobCount(self, filter=None):
        query_string = ""
        if isinstance(filter, str) and filter is not None:
            if len(query_string) == 0:
                query_string = f"?filter={filter}"
            else:
                query_string += f"&filter={filter}"
        response = requests.get(f"{self.mcrit_server}/jobs{query_string}", headers=self.headers)
        if self.raw:
            return response
        data = handle_response(response)
        if data is not None:
            return len(data)

    def getQueueData(self, start=0, limit=0, filter=None):
        """
        Get queue data, optionally from <start> and <limit> many
        Supported by mcritweb API pass-through
        """
        query_string = ""
        if isinstance(start, int) and start > 0:
            if len(query_string) == 0:
                query_string = f"?start={start}"
            else:
                query_string += f"&start={start}"
        if isinstance(limit, int) and limit > 0:
            if len(query_string) == 0:
                query_string = f"?limit={limit}"
            else:
                query_string += f"&limit={limit}"
        if isinstance(filter, str) and filter is not None:
            if len(query_string) == 0:
                query_string = f"?filter={filter}"
            else:
                query_string += f"&filter={filter}"
        response = requests.get(f"{self.mcrit_server}/jobs/{query_string}", headers=self.headers)
        if self.raw:
            return response
        data = handle_response(response)
        if data is not None:
            return [Job(job_data, None) for job_data in data]

    def getJobData(self, job_id):
        """
        Get the Job for a given <job_id>
        Supported by mcritweb API pass-through
        """
        response = requests.get(f"{self.mcrit_server}/jobs/{job_id}", headers=self.headers)
        if self.raw:
            return response
        data = handle_response(response)
        if data is not None:
            return Job(data, None)

    def getResultForJob(self, job_id):
        """
        Get the Result for Job with a given <job_id>
        Supported by mcritweb API pass-through
        """
        response = requests.get(f"{self.mcrit_server}/jobs/{job_id}/result", headers=self.headers)
        if self.raw:
            return response
        return handle_response(response)

    def getResult(self, result_id):
        """
        Get the Result for a given <result_id>
        Supported by mcritweb API pass-through
        """
        response = requests.get(f"{self.mcrit_server}/results/{result_id}", headers=self.headers)
        if self.raw:
            return response
        return handle_response(response)

    def getJobForResult(self, result_id):
        """
        Get the Job for the Result with a given <result_id>
        Supported by mcritweb API pass-through
        """
        response = requests.get(f"{self.mcrit_server}/results/{result_id}/job", headers=self.headers)
        if self.raw:
            return response
        data = handle_response(response)
        if data is not None:
            return Job(data, None)

    def awaitResult(self, job_id, sleep_time=2):
        if job_id is None:
            return None
        job = self.getJobData(job_id)
        while not isJobFinishedTerminatedOrFailed(job):
            time.sleep(sleep_time)
            job = self.getJobData(job_id)
        if isJobTerminated(job):
            raise JobTerminatedError
        result_id = job.result
        return self.getResult(result_id)

    ###########################################
    ### Import / Export
    ###########################################

    def getExportData(self, sample_ids=None, compress_data=True) -> dict:
        compress_uri_param = "?compress=True" if compress_data else ""
        result_data = {}
        if sample_ids is not None:
            if isinstance(sample_ids, list) and all(isinstance(item, int) for item in sample_ids):
                sample_ids_as_str = ",".join([str(sample_id) for sample_id in sample_ids])
                response = requests.get(f"{self.mcrit_server}/export/{sample_ids_as_str}{compress_uri_param}", headers=self.headers)
                result_data = handle_response(response)
            else:
                raise ValueError("sample_ids must be a list of int.")
        else:
            response = requests.get(f"{self.mcrit_server}/export{compress_uri_param}", headers=self.headers)
            result_data = handle_response(response)
        return result_data

    def addImportData(self, import_data):
        if not isinstance(import_data, dict):
            raise ValueError("Can only forward dictionaries with export data.")
        response = requests.post(f"{self.mcrit_server}/import", json=import_data, headers=self.headers)
        return handle_response(response)

    ###########################################
    ### Unique Blocks
    ###########################################

    def requestUniqueBlocksForSamples(self, sample_ids: List[int]) -> Dict:
        if isinstance(sample_ids, list) and all(isinstance(item, int) for item in sample_ids):
            sample_ids_as_str = ",".join([str(sample_id) for sample_id in sample_ids])
            response = requests.get(f"{self.mcrit_server}/uniqueblocks/samples/{sample_ids_as_str}", headers=self.headers)
            result_data = handle_response(response)
        else:
            raise ValueError("sample_ids must be a list of int.")
        return result_data

    def requestUniqueBlocksForFamily(self, family_id: int) -> Dict:
        if isinstance(family_id, int):
            response = requests.get(f"{self.mcrit_server}/uniqueblocks/family/{family_id}", headers=self.headers)
            result_data = handle_response(response)
        else:
            raise ValueError("sample_ids must be a list of int.")
        return result_data

    ###########################################
    ### Search
    ###########################################

    # When performing an initial search, the cursor should be set to None.
    # Search results are of the following form:
    # {
    #     "search_results": {
    #         id1: found_entry1,
    #         id2: found_entry2,
    #         ...
    #     },
    #     "cursor": {
    #         "forward": forward cursor,
    #         "backward": backward cursor,
    #     }
    # }
    # To get further results, perform a search using the forward cursor.
    # To get back to the previous search results, use the backward cursor.
    # If no further or previous results are available, the forward or backward cursor will be None.
    #
    # IMPORTANT: A cursor shall only be used in combination with the same
    # search_term, is_ascending and sort_by value that were used when the cursor was returned from mcrit.
    # If those parameters are altered, mcrit's behavior is undefined.
    
    def _search_base(self, search_kind, search_term, cursor=None, is_ascending=True, sort_by=None, limit=None):
        params = {
            "query": search_term,
            "is_ascending": is_ascending,
        }
        if cursor is not None:
            params["cursor"] = cursor
        if sort_by is not None:
            params["sort_by"] = sort_by 
        if limit is not None:
            params["limit"] = limit 
        encoded_params = urllib.parse.urlencode(params)
        response = requests.get(f"{self.mcrit_server}/search/{search_kind}?{encoded_params}", headers=self.headers)
        return handle_response(response)

    search_families = functools.partialmethod(_search_base, "families")

    search_samples = functools.partialmethod(_search_base, "samples")

    search_functions = functools.partialmethod(_search_base, "functions")



import json
import logging
import os

import falcon

from mcrit.index.MinHashIndex import MinHashIndex
from mcrit.server.FamilyResource import FamilyResource
from .FunctionResource import FunctionResource
from .JobResource import JobResource
from .MatchResource import MatchResource
from .BlocksResource import BlocksResource
from .QueryResource import QueryResource
from .SampleResource import SampleResource
from .StatusResource import StatusResource

# Only do basicConfig if no handlers have been configured
if len(logging._handlerList) == 0:
    logging.basicConfig(level=logging.INFO, format="%(asctime)-15s %(message)s")
LOGGER = logging.getLogger(__name__)



def create_index():
    # TODO we will want to load config values and everyting from the database instead of initializing everytime here
    index = MinHashIndex()
    this_path = str(os.path.abspath(__file__))
    root_path = os.sep.join(this_path.split(os.sep)[:-2])
    index_path = os.sep.join([root_path, "db", "index.json"])
    index_json = None
    if os.path.isfile(index_path):
        LOGGER.info("Found index.json, loading and using data...")
        with open(index_path, "r") as fjson:
            index_json = json.load(fjson)
        index.setStorageData(index_json)
    return index


def get_app():
    index = create_index()
    status_resource = StatusResource(index)
    family_resource = FamilyResource(index)
    sample_resource = SampleResource(index)
    function_resource = FunctionResource(index)
    match_resource = MatchResource(index)
    block_resource = BlocksResource(index)
    query_resource = QueryResource(index)
    job_resource = JobResource(index)

    _app = falcon.App()
    _app.req_options.strip_url_path_trailing_slash = True
    _app.add_route("/", status_resource)
    _app.add_route("/status", status_resource, suffix="status")
    _app.add_route("/version", status_resource, suffix="version")
    _app.add_route("/config", status_resource, suffix="config")
    _app.add_route("/export", status_resource, suffix="export")
    _app.add_route("/export/{comma_separated_sample_ids}", status_resource, suffix="export_selection")
    # NOTE: adds to storage, does not replace storage
    _app.add_route("/import", status_resource, suffix="import")  # post
    # drops storage and sets up new empty instance
    _app.add_route("/respawn", status_resource, suffix="respawn")  # post
    # schedule a job that calculates all missing minhashes for all samples/functions, in case the respective jobs failed before
    _app.add_route("/complete_minhashes", status_resource, suffix="complete_minhashes")  # get
        # schedule a job that drops all minhash bands and rebuilds the index using all existing minhashes
    _app.add_route("/rebuild_index", status_resource, suffix="rebuild_index")  # get

    # search suitable fields based on query
    _app.add_route("/search/families", status_resource, suffix="search_families")
    _app.add_route("/search/samples", status_resource, suffix="search_samples")
    _app.add_route("/search/functions", status_resource, suffix="search_functions")

    _app.add_route("/families", family_resource, suffix="collection")
    # supports GET, PUT (for modification of family_name, is_library), DELETE (for all samples)
    _app.add_route("/families/{family_id:int}", family_resource)

    # supports GET and POST (to insert one sample)
    _app.add_route("/samples", sample_resource, suffix="collection")
    # post only
    _app.add_route("/samples/binary", sample_resource, suffix="submit_binary")
    # supports GET, PUT (for modification of family_name, version, component, is_library), DELETE (for one sample)
    _app.add_route("/samples/{sample_id:int}", sample_resource)
    #
    _app.add_route("/samples/sha256/{sample_sha256}", sample_resource, suffix="by_sha256")
    _app.add_route("/samples/{sample_id:int}/functions", sample_resource, suffix="functions")
    _app.add_route(
        "/samples/{sample_id:int}/functions/{function_id:int}",
        sample_resource,
        suffix="function",
    )

    _app.add_route("/functions", function_resource, suffix="collection")
    _app.add_route("/functions/{function_id:int}", function_resource)

    _app.add_route("/matches/sample/{sample_id:int}", match_resource, suffix="sample")
    _app.add_route("/matches/sample/cross/{sample_ids}", match_resource, suffix="sample_cross")
    _app.add_route("/matches/sample/{sample_id:int}/{sample_id_b:int}", match_resource, suffix="sample_vs")
    _app.add_route("/matches/function/{function_id:int}/{function_id_b:int}", match_resource, suffix="function_vs")
    _app.add_route("/matches/function/{function_id:int}", match_resource, suffix="function")

    _app.add_route("/uniqueblocks/samples/{comma_separated_sample_ids}", block_resource, suffix="unique_blocks_for_samples")
    _app.add_route("/uniqueblocks/family/{family_id:int}", block_resource, suffix="unique_blocks_for_family")

    _app.add_route("/query", query_resource, suffix="query_smda")
    _app.add_route("/query/binary", query_resource, suffix="query_binary")
    _app.add_route(
        "/query/binary/mapped/{base_address}",
        query_resource,
        suffix="query_binary_mapped",
    )
    _app.add_route("/query/function", query_resource, suffix="query_smda_function")
    _app.add_route("/query/pichash/{pichash}", query_resource, suffix="query_pichash")
    _app.add_route("/query/pichash/{pichash}/summary", query_resource, suffix="query_pichash_summary")
    _app.add_route("/query/picblockhash/{picblockhash}", query_resource, suffix="query_picblockhash")
    _app.add_route("/query/picblockhash/{picblockhash}/summary", query_resource, suffix="query_picblockhash_summary")

    _app.add_route("/jobs", job_resource, suffix="collection")
    _app.add_route("/jobs/{job_id}", job_resource)
    _app.add_route("/jobs/{job_id}/result", job_resource, suffix="job_result")
    _app.add_route("/results/{result_id}", job_resource, suffix="results")
    _app.add_route("/results/{result_id}/job", job_resource, suffix="result_job")

    return _app



import re
import logging

import falcon

from mcrit.server.utils import timing, jsonify
from mcrit.index.MinHashIndex import MinHashIndex
from mcrit.server.utils import db_log_msg

# TODO these should also return status and data in their json response

class JobResource:
    def __init__(self, index: MinHashIndex):
        self.index = index

    @timing
    def on_get_collection(self, req, resp):
        # parse optional request parameters
        query_filter = None
        if "filter" in req.params:
            query_filter = req.params["filter"]
        start_job_id = 0 
        if "start" in req.params:
            try:
                start_job_id = int(req.params["start"])
            except:
                pass
        limit_job_count = 0 
        if "limit" in req.params:
            try:
                limit_job_count = int(req.params["limit"])
            except:
                pass
        # newest first
        queue_data = reversed(self.index.getQueueData(filter=query_filter))
        result_data = []
        num_jobs_included = 0
        for index, job_data in enumerate(queue_data):
            if index < start_job_id:
                continue
            if (limit_job_count == 0) or  (num_jobs_included < limit_job_count):
                result_data.append(job_data)
                num_jobs_included += 1
            else:
                break
            if limit_job_count and num_jobs_included >= limit_job_count:
                break
        resp.data = jsonify({"status": "successful", "data": result_data})
        db_log_msg(self.index, req, f"JobResource.on_get_collection - success.")

    @timing
    def on_get(self, req, resp, job_id=None):
        # validate that we only allow hexstrings with 24 chars
        if not re.match("[a-fA-F0-9]{24}", job_id):
            resp.status = falcon.HTTP_400
            resp.data = jsonify({"status": "failed", "data": {"message": "Valid JobIDs are hexstrings with 24 characters."}})
            db_log_msg(self.index, req, f"JobResource.on_get - failed - invalid job_id.")
            return  
        data = self.index.getJobData(job_id)
        # TODO throw 404 if job_id is unknown
        # resp.status = falcon.HTTP_404
        resp.data = jsonify({"status": "successful", "data": data})
        db_log_msg(self.index, req, f"JobResource.on_get - success.")

    @timing
    def on_get_results(self, req, resp, result_id=None):
        # validate that we only allow hexstrings with 24 chars
        if not re.match("[a-fA-F0-9]{24}", result_id):
            resp.status = falcon.HTTP_400
            resp.data = jsonify({"status": "failed", "data": {"message": "Valid ResultIDs are hexstrings with 24 characters."}})
            db_log_msg(self.index, req, f"JobResource.on_get_results - failed - invalid result_id.")
            return  
        data = self.index.getResult(result_id)
        # TODO throw 404 if job_id is unknown
        # resp.status = falcon.HTTP_404
        resp.data = jsonify({"status": "successful", "data": data})
        db_log_msg(self.index, req, f"JobResource.on_get_results - success.")

    @timing
    def on_get_job_result(self, req, resp, job_id=None):
        # validate that we only allow hexstrings with 24 chars
        if not re.match("[a-fA-F0-9]{24}", job_id):
            resp.status = falcon.HTTP_400
            resp.data = jsonify()
            db_log_msg(self.index, req, f"JobResource.on_get_job_result - failed - invalid job_id.")
            return  
        data = self.index.getResultForJob(job_id)
        # TODO throw 404 if job_id is unknown
        # resp.status = falcon.HTTP_404
        resp.data = jsonify({"status": "successful", "data": data})
        db_log_msg(self.index, req, f"JobResource.on_get_job_result - success.")

    @timing
    def on_get_result_job(self, req, resp, result_id=None):
        # validate that we only allow hexstrings with 24 chars
        if not re.match("[a-fA-F0-9]{24}", result_id):
            resp.status = falcon.HTTP_400
            resp.data = jsonify({"status": "failed", "data": {"message": "Valid ResultIDs are hexstrings with 24 characters."}})
            db_log_msg(self.index, req, f"JobResource.on_get_job_result - failed - invalid result_id.")
            return  
        job_id = self.index.getJobIdForResult(result_id)
        data = self.index.getJobData(job_id)
        # TODO throw 404 if job_id is unknown
        # resp.status = falcon.HTTP_404
        resp.data = jsonify({"status": "successful", "data": data})
        db_log_msg(self.index, req, f"JobResource.on_get_result_job - success.")


import struct

import numpy as np

try:
    # try with a fast c-implementation ...
    import mmh3
except ImportError:
    # ... otherwise fallback to the pure python version
    import mcrit.libs.pymmh3 as mmh3


class MinHash(object):
    """DTO for an actual MinHash
    <minhash>: a binary sequence of packed int8/32 values
    <minhash_int>: the equivalent representation of <minhash> but as list of int8/32
    """

    _HASH_MAX = 0xFFFFFFFF
    _MINHASH_BITS = 32

    def __init__(self, function_id=None, minhash_bytes=None, minhash_signature=None, minhash_bits=32):
        self.minhash = b""
        self.minhash_int = []
        if minhash_bits:
            self._MINHASH_BITS = minhash_bits
        if minhash_bytes and minhash_signature:
            raise ValueError("Can use only one keyword argument")
        if minhash_bytes:
            if self._MINHASH_BITS <= 8:
                minhash_signature = np.frombuffer(minhash_bytes, dtype=np.uint8)
            else:
                minhash_signature = np.frombuffer(minhash_bytes, dtype=np.uint32)
            self.setMinHash(minhash_signature)
        elif minhash_signature:
            self.setMinHash(minhash_signature)

        self.shingler_composition = {}
        self.function_id = function_id

    def hasMinHash(self) -> bool:
        return len(self.minhash) > 0

    def getMinHash(self):
        return self.minhash

    def getMinHashInt(self):
        return self.minhash_int

    def getSignatureEntrySize(self):
        return 1 if MinHash._MINHASH_BITS <= 8 else 4

    def setMinHash(self, minhash_signature):
        self.minhash_int = [i % 2 ** self._MINHASH_BITS for i in minhash_signature]
        if self._MINHASH_BITS <= 8:
            self.minhash = np.array(self.minhash_int, dtype=np.uint8).tobytes()
        else:
            self.minhash = np.array(self.minhash_int, dtype=np.uint32).tobytes()

    def getComposition(self):
        return self.shingler_composition

    def scoreAgainst(self, other):
        return MinHash.calculateMinHashScore(self.minhash, other.minhash, minhash_bits=self._MINHASH_BITS)

    def __str__(self) -> str:
        return "Function ID: {}, Minhash: {}...".format(self.function_id, self.minhash[:16])

    @staticmethod
    def getHashMax():
        return MinHash._HASH_MAX

    @staticmethod
    def hashData(data, seed) -> int:
        if isinstance(data, (str, bytes, bytearray)):
            return mmh3.hash(data, seed) & MinHash._HASH_MAX
        elif isinstance(data, list):
            to_hash = "".join([str(elem) for elem in data])
            return mmh3.hash(to_hash, seed) & MinHash._HASH_MAX
        else:
            raise NotImplementedError

    @staticmethod
    def calculateMinHashScore(first, second, minhash_bits=32):
        if minhash_bits <= 8:
            first_np = np.frombuffer(first, dtype=np.uint8)
            second_np = np.frombuffer(second, dtype=np.uint8)
        else:
            first_np = np.frombuffer(first, dtype=np.uint32)
            second_np = np.frombuffer(second, dtype=np.uint32)
        return 100.0 * sum(first_np == second_np) / len(first_np)

    @staticmethod
    def calculateMinHashIntScore(first, second):
        score = 0
        num_hashes = len(first)
        if num_hashes:
            for index, part in enumerate(first):
                score += 1 if part == second[index] else 0
            return 100.0 * score / num_hashes
        return 0.0

    def __repr__(self):
        # return f"<Minhash(function_id={self.function_id}, length={len(self.minhash_int)})>"
        return "<Minhash(function_id={}, length={})>".format(self.function_id, len(self.minhash_int))



import logging

import falcon
import re
import json

from mcrit.server.utils import timing, jsonify
from mcrit.index.MinHashIndex import MinHashIndex
from mcrit.server.utils import db_log_msg


class SampleResource:
    def __init__(self, index: MinHashIndex):
        self.index = index

    @timing
    def on_get(self, req, resp, sample_id=None):
        if not self.index.isSampleId(sample_id):
            resp.data = jsonify(
                {
                    "status": "failed",
                    "data": {"message": "We don't have a sample with that id."},
                }
            )
            resp.status = falcon.HTTP_404
            db_log_msg(self.index, req, f"SampleResource.on_get - failed - unknown sample_id: {sample_id}.")
            return
        resp.data = jsonify({"status": "successful", "data": self.index.getSampleById(sample_id).toDict()})
        db_log_msg(self.index, req, f"SampleResource.on_get - success.")

    @timing
    def on_get_by_sha256(self, req, resp, sample_sha256):
        sha256_pattern = "[a-fA-F0-9]{64}"
        match = re.match(sha256_pattern, sample_sha256)
        if not match:
            resp.data = jsonify(
                {
                    "status": "failed",
                    "data": {"message": "No valid SHA256 provided."},
                }
            )
            resp.status = falcon.HTTP_400
            db_log_msg(self.index, req, f"SampleResource.on_get_by_sha256 - failed - invalid SHA256.")
            return
        sample_entry = self.index.getSampleBySha256(sample_sha256)
        if sample_entry is None:
            resp.data = jsonify(
                {
                    "status": "failed",
                    "data": {"message": "We don't have a sample with that SHA256."},
                }
            )
            resp.status = falcon.HTTP_404
            db_log_msg(self.index, req, f"SampleResource.on_get_by_sha256 - failed - unknown SHA256: {sample_sha256}.")
            return

        resp.data = jsonify({"status": "successful", "data": sample_entry.toDict()})
        db_log_msg(self.index, req, f"SampleResource.on_get_by_sha256 - success.")

    @timing
    def on_delete(self, req, resp, sample_id=None):
        successful = self.index.deleteSample(sample_id, force_recalculation=True)
        if successful:
            resp.data = jsonify({"status": "successful", "data": successful})
            resp.status = falcon.HTTP_202
            db_log_msg(self.index, req, f"SampleResource.on_delete - success.")
        else:
            resp.data = jsonify({"status": "failed", "data": {"message": "Failed to delete sample."}})
            # TODO whats the correct code?
            resp.status = falcon.HTTP_400
            db_log_msg(self.index, req, f"SampleResource.on_delete - failed - sample_id: {sample_id} not deleted.")

    @timing
    def on_put(self, req, resp, sample_id=None):
        resp.status = falcon.HTTP_400
        if not req.content_length or not isinstance(req.media, dict):
            resp.data = jsonify({"status": "failed","data": {"message": "PUT request without body can't be processed."}})
            db_log_msg(self.index, req, f"SampleResource.on_put - failed - no PUT body.")
            return
        # sanitize sample information
        information_update = req.media
        if "family_name" in information_update and not re.match("^(?=[a-zA-Z0-9._\-]{0,64}$)(?!.*[\-_.]{2})[^\-_.].*[^\-_.]$", information_update["family_name"]):
            resp.data = jsonify({"status": "failed","data": {"message": "family_name may be 0-64 alphanumeric chars with single dots, dashes, underscores inbetween."}})
            db_log_msg(self.index, req, f"SampleResource.on_put - failed - invalid family name.")
            return
        if "version" in information_update and not re.match("^[ -~]{1,64}$", information_update["version"]):
            resp.data = jsonify({"status": "failed","data": {"message": "version may be 0-64 printable characters."}})
            db_log_msg(self.index, req, f"SampleResource.on_put - failed - invalid version name.")
            return
        if "component" in information_update and not re.match("^[ -~]{1,64}$", information_update["component"]):
            resp.data = jsonify({"status": "failed","data": {"message": "component may be 0-64 printable characters."}})
            db_log_msg(self.index, req, f"SampleResource.on_put - failed - invalid component name.")
            return
        if "is_library" in information_update:
            if not (isinstance(information_update["is_library"], bool) or information_update["is_library"] in ["True", "False", "true", "false", "0", "1", 0, 1]):
                resp.data = jsonify({"status": "failed","data": {"message": "is_library must be boolean."}})
                db_log_msg(self.index, req, f"SampleResource.on_put - failed - is_library must be boolean.")
                return
            if information_update["is_library"] in ["True", "true", "1", 1]:
                information_update["is_library"] = True
            elif information_update["is_library"] in ["False", "false", "0", 0]:
                information_update["is_library"] = False
        successful = self.index.modifySample(sample_id, information_update, force_recalculation=True)
        if successful:
            resp.data = jsonify({"status": "successful", "data": {"message": "Sample modified."}})
            resp.status = falcon.HTTP_202
            db_log_msg(self.index, req, f"SampleResource.on_put - success - modified sample_id {sample_id}.")
        else:
            resp.data = jsonify({"status": "failed", "data": {"message": "Failed to modify sample."}})
            db_log_msg(self.index, req, f"SampleResource.on_put - failed - sample_id {sample_id} not modified.")

    @timing
    def on_post_collection(self, req, resp):
        # TODO 2019-05-07 verify integrity of SMDA report
        if not req.content_length:
            resp.data = jsonify(
                {
                    "status": "failed",
                    "data": {"message": "POST request without body can't be processed."},
                }
            )
            resp.status = falcon.HTTP_400
            db_log_msg(self.index, req, f"SampleResource.on_post_collection - failed - JSON not processible.")
            return
        username = req.get_header("username", default="anonymous")
        summary = self.index.addReportJson(req.media, username=username)
        if summary is not None:
            resp.data = jsonify({"status": "successful", "data": summary})
            db_log_msg(self.index, req, f"SampleResource.on_post_collection - success.")
        else:
            resp.data = jsonify({"status": "failed", "data": {"message": "Could not process JSON."}})
            db_log_msg(self.index, req, f"SampleResource.on_post_collection - failed - JSON not processible.")

    @timing
    def on_post_submit_binary(self, req, resp):
        if not req.content_length:
            resp.data = jsonify(
                {
                    "status": "failed",
                    "data": {"message": "POST request without body can't be processed."},
                }
            )
            resp.status = falcon.HTTP_400
            db_log_msg(self.index, req, f"SampleResource.on_post_submit_binary - failed - no POST body.")
            return
        # TODO parse respective query fields -> escape / sanitize input
        filename = req.params["filename"] if "filename" in req.params else None
        family = req.params["family"] if "family" in req.params else None
        version = req.params["version"] if "version" in req.params else None
        is_dump = True if ("is_dump" in req.params and req.params["is_dump"] in ["True", "true", "1", 1]) else False
        base_address = 0 
        if "base_addr" in req.params:
            try:
                base_address = int(req.params["base_addr"], 16)
            except:
                pass
        bitness = int(req.params["bitness"]) if ("bitness" in req.params and req.params["bitness"] in ["32", "64"]) else 32
        # binary itself
        binary = req.stream.read()
        job_id = self.index.addBinarySample(binary, filename, family, version, is_dump, base_address, bitness)
        # TODO 2019-05-10 return full sample_entry in response
        resp.data = jsonify({"status": "successful", "data": job_id})
        db_log_msg(self.index, req, f"SampleResource.on_post_submit_binary - success - with job_id: {job_id}.")

    @timing
    def on_get_collection(self, req, resp):
        # parse optional request parameters
        start_index = 0 
        if "start" in req.params:
            try:
                start_index = int(req.params["start"])
            except:
                pass
        limit_sample_count = 0 
        if "limit" in req.params:
            try:
                limit_sample_count = int(req.params["limit"])
            except:
                pass
        sample_overview = {}
        sample_entries = self.index.getSamples(start_index, limit_sample_count)
        for sample_entry in sample_entries:
            sample_overview[sample_entry.sample_id] = sample_entry.toDict()
        resp.data = jsonify({"status": "successful", "data": sample_overview})
        db_log_msg(self.index, req, f"SampleResource.on_get_collection - success.")

    @timing
    def on_get_function(self, req, resp, sample_id=None, function_id=None):
        if not self.index.isSampleId(sample_id):
            resp.data = jsonify({"status": "failed", "data": {"message": "We don't have a sample with that id."}})
            resp.status = falcon.HTTP_404
            db_log_msg(self.index, req, f"SampleResource.on_get_function - failed - unknown sample_id {sample_id}.")
            return
        if not self.index.isFunctionId(function_id):
            resp.data = jsonify({"status": "failed", "data": {"message": "We don't have a function with that id."}})
            resp.status = falcon.HTTP_404
            db_log_msg(self.index, req, f"SampleResource.on_get_function - failed - unknown function_id {function_id}.")
            return
        data = self.index.getFunctionById(function_id).toDict()
        # check if function_id belongs to sample_id
        if data["sample_id"] != sample_id:
            resp.data = jsonify({"status": "failed", "data": {"message": "This sample doesn't have a function with that id."}})
            resp.status = falcon.HTTP_404
            db_log_msg(self.index, req, f"SampleResource.on_get_function - failed - unknown function_id {function_id} for sample_id {sample_id}.")
            return
        resp.data = jsonify({"status": "successful", "data": data})
        db_log_msg(self.index, req, f"SampleResource.on_get_function - success.")

    @timing
    def on_get_functions(self, req, resp, sample_id=None):
        if not self.index.isSampleId(sample_id):
            resp.data = jsonify(
                {
                    "status": "failed",
                    "data": {"message": "We don't have a sample with that id."},
                }
            )
            resp.status = falcon.HTTP_404
            db_log_msg(self.index, req, f"SampleResource.on_get_functions - failed - unknown sample_id {sample_id}.")
            return
        function_entries = self.index.getFunctionsBySampleId(sample_id)

        data = {
            function_entry.function_id: function_entry.toDict() 
            for function_entry 
            in sorted(function_entries, key=lambda func: func.offset)
        }
        resp.data = jsonify(
            {
                "status": "successful",
                "data": data
            }
        )
        db_log_msg(self.index, req, f"SampleResource.on_get_functions - success.")



#   Copyright 2012 Kapil Thangavelu
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


import json
import traceback
from datetime import datetime, timedelta
from typing import List

import gridfs
import pymongo
from pymongo import MongoClient, ReturnDocument
from bson.objectid import ObjectId

DEFAULT_INSERT = {
    "locked_by": None,
    "locked_at": None,
    "last_error": None,
    "finished_at": None,
    "created_at": None,
    "started_at": None,
    "terminated": False,
    "required_by": [],
    "number": None,
    "result": None,
    "notify_done": False,
}

# added to track incremental job numbers via MCRIT storage (apart from OIDs)
def _useCounter(database, name: str) -> int:
    result = database.counters.find_one_and_update(
        filter={"name": name}, 
        update={"$inc": {"value": 1}}, 
        upsert=True
    )
    if result is None:
        return 0
    return result["value"]


class MongoQueue(object):
    """A queue class"""

    def __init__(self, queue_config, consumer_id, timeout=300, max_attempts=3):
        """ """
        self.queue_config = queue_config
        self.collection = None
        self.consumer_id = consumer_id
        self.timeout = timeout
        self.max_attempts = max_attempts
        self._default_insert = dict(DEFAULT_INSERT)
        self._default_insert["attempts_left"] = max_attempts
        self.fs = None
        self.fs_files = None
        self.cache_time = 10 ** 9

    def _getCollection(self):
        # because of gunicorn and forking workers, we want to delay creation of MongoClient until actual usage and avoid it within __init__()
        if self.collection is None:
            db = MongoClient(host=self.queue_config.QUEUE_SERVER, port=self.queue_config.QUEUE_PORT, connect=False)
            self.collection = db[self.queue_config.QUEUE_MONGODB_DBNAME][self.queue_config.QUEUE_MONGODB_COLLECTION_NAME]
            self.fs = gridfs.GridFS(self.collection.database)
            self.fs_files = self.collection.database["fs.files"]
            self._ensure_indices()
        return self.collection
    
    def _getFs(self):
        if self.fs is None:
            collection = self._getCollection()
            self.fs = gridfs.GridFS(collection.database)
            self.fs_files = self.collection.database["fs.files"]
        return self.fs
    
    def _getFsFiles(self):
        if self.fs_files is None:
            collection = self._getCollection()
            self.fs = gridfs.GridFS(self.collection.database)
            self.fs_files = collection.database["fs.files"]
        return self.fs_files

    def _ensure_indices(self):
        # should only be called, after self.collection has been initiated
        self.collection.create_index("payload.descriptor")
        self.fs_files.create_index("metadata.sha256")

    def close(self):
        """Close the in memory queue connection."""
        self._getCollection().connection.close()

    def clear(self):
        """Clear the queue."""
        self._getCollection().database["fs"].drop()
        self._getCollection().drop()
        self._ensure_indices()

    def size(self):
        """Total size of the queue"""
        return self._getCollection().count_documents({"finished_at": None})

    def size_inc_finished(self):
        """Total size of the queue"""
        return self._getCollection().count_documents(ilter={})

    def repair(self):
        """Clear out stale locks.

        Increments per job attempt counter.
        """
        self._getCollection().find_one_and_update(
            filter={"locked_by": {"$ne": None}, "locked_at": {"$lt": datetime.now() - timedelta(self.timeout)}},
            update={"$set": {"locked_by": None, "locked_at": None}, "$inc": {"attempts_left": -1}},
        )

    def drop_max_attempts(self):
        """ """
        self._getCollection().find_one_and_update({"attempts_left": {"$lte": 0}}, remove=True)

    def put(self, payload, priority=0, await_jobs:List[str]=[]):
        """Place a job into the queue"""
        job = dict(self._default_insert)
        job["number"] = _useCounter(self._getCollection().database, "job")
        job["created_at"] = datetime.now()
        job["priority"] = priority
        job["payload"] = payload
        await_jobs_set = set(await_jobs)
        job["unfinished_dependencies"] = list(await_jobs_set)
        job["all_dependencies"] = list(await_jobs_set)
        insert_result = self._getCollection().insert_one(job)
        if insert_result.acknowledged:
            job_id = insert_result.inserted_id
            for child_job_id in await_jobs_set:
                self._notify_on_done(child_job_id, job_id)
            return job_id
        return None

    def _notify_on_done(self, notifying_job_id:str, notified_job_id:str):
        self._getCollection().find_one_and_update(
            filter={"_id": ObjectId(notifying_job_id)},
            update={"$push": {'required_by': notified_job_id}}
        )
        notifying_job = self._getCollection().find_one({"_id": ObjectId(notifying_job_id)})
        if notifying_job["notify_done"]:
            self._getCollection().find_one_and_update(
                filter={"_id": ObjectId(notified_job_id)},
                update={"$pull": {'unfinished_dependencies': notifying_job_id}},
            )
        #     # atomically
        #     # should handle already deleted deps gracefully
        #     remove job.id from job_to_notify.unfinished_deps
    
    def _notify_dependent_jobs(self, job_id:str):
        job = self._getCollection().find_one_and_update(
            filter={"_id": ObjectId(job_id)},
            update={"$set": {'notify_done': True}},
        )
        for job_id_to_notify in job["required_by"]:
            self._getCollection().find_one_and_update(
                filter={"_id": ObjectId(job_id_to_notify)},
                update={"$pull": {'unfinished_dependencies': job_id}},
            )

    def next(self):
        current_time = datetime.now()
        return self._wrap_one(
            self._getCollection().find_one_and_update(
                filter={
                    "locked_by": None,
                    "locked_at": None,
                    "attempts_left": {"$gt": 0},
                    "finished_at": None,
                    "unfinished_dependencies": [],
                },
                update={"$set": {"locked_by": self.consumer_id, "locked_at": current_time, "started_at": current_time}},
                sort=[("priority", pymongo.DESCENDING), ("created_at", pymongo.ASCENDING)],
                new=1,
                # limit=1
            )
        )

    def _jobs_to_do(self):
        return self._getCollection().find(
            filter={"locked_by": None, "locked_at": None, "attempts_left": {"$gt": 0}, "finished_at": None},
            sort=[("priority", pymongo.DESCENDING)],
        )

    def _jobs_in_progress(self):
        return self._getCollection().find(
            filter={
                "locked_by": {"$ne": None},
                "locked_at": {"$ne": None},
                "attempts_left": {"gt": 0},
                "finished_at": None,
            },
            sort=[("priority", pymongo.DESCENDING)],
        )

    def _jobs_given_up(self):
        return self._getCollection().find(
            filter={"attempts_left": {"$le": 0}, "finished_at": None},
            sort=[("priority", pymongo.DESCENDING)],
        )

    def _jobs_finished(self):
        return self._getCollection().findOne(
            filter={"finished_at": {"$ne": None}},
            sort=[("priority", pymongo.DESCENDING)],
        )

    def _wrap_one(self, data):
        return data and Job(self, data) or None

    def stats(self):
        """Get statistics on the queue.

        Use sparingly requires a collection lock.
        """
        queries = []
        queries.append({"locked_by": None, "attempts_left": {"$gt": 0}})
        queries.append({"locked_by": {"$ne": None}})
        queries.append({"attempts_left": {"$lte": 0}})
        queries.append({})
        counts = [self._getCollection().count_documents(q) for q in queries]

        return dict(zip(["available", "locked", "errors", "total"], counts))

    def get_jobs(self):
        jobs = []
        for job_document in self._getCollection().find():
            jobs.append(self._wrap_one(job_document))
        return jobs

    def get_job(self, job_id):
        job_id = ObjectId(job_id)
        return self._wrap_one(self._getCollection().find_one({"_id": job_id}))

    def _file_to_grid(self, binary, metadata=None):
        object_id = self._getFs().put(binary, metadata=metadata)
        return str(object_id)

    def _grid_to_file(self, grid, results_only=True):
        oid = ObjectId(grid)
        if not self._getFs().exists(oid):
            return None
        entry = self._getFs().get(oid)
        if results_only:
            metadata = entry.metadata
            if not "result" in metadata or not metadata["result"]:
                return b'"Access Not Allowed"'
        result = entry.read()
        return result

    def _dicts_to_grid(self, dicts, **kwargs):
        return self._file_to_grid(json.dumps(dicts).encode("ascii"), **kwargs)

    def _grid_to_dicts(self, grid, **kwargs):
        result = None
        grid_file = self._grid_to_file(grid, **kwargs)
        if grid_file is not None:
            decoded_file = grid_file.decode("ascii")
            result = json.loads(grid_file)
        return result

    def _delete_grid(self, grid):
        oid = ObjectId(grid)
        self._getFs().delete(oid)

    def _grid_to_meta(self, grid):
        oid = ObjectId(grid)
        entry = self._getFs().get(oid)
        return entry.metadata

    def get_cached_job_id(self, payload):
        job = self._wrap_one(
            self._getCollection().find_one(
                {
                    "attempts_left": {"$gt": 0},
                    "payload.descriptor": payload["descriptor"],
                    "terminated": False,
                },
                sort=[("created_at", pymongo.DESCENDING)],
            )
        )
        return job and job.job_id or None

    def _get_file_by_hash(self, sha256):
        file = self._getFs().find_one({"metadata.sha256": sha256})

    def get_file_by_hash_inc_lock(self, sha256):
        file = self._getFsFiles().find_one_and_update({"metadata.sha256": sha256}, {"$inc": {"metadata.tmp_lock": 1}})
        return file and str(file["_id"]) or None

    def add_job_id_to_file(self, job_id, file_id):
        file_id = ObjectId(file_id)
        self._getFsFiles().find_one_and_update(
            {"_id": file_id}, {"$inc": {"metadata.tmp_lock": -1}, "$addToSet": {"metadata.jobs": job_id}}
        )

    def clean(self):
        time_threshold = datetime.now() - timedelta(seconds=self.cache_time)
        job_query = {"finished_at": {"$lt": time_threshold}}
        to_delete = self._getCollection().find(job_query)
        to_delete = list(to_delete)
        results = [data["result"] for data in to_delete]
        file_params = {data["_id"]: list(json.loads(data["payload"]["file_params"]).values()) for data in to_delete}

        # delete results
        for r in results:
            self._delete_grid(r)

        # remove job from params
        for job, params in file_params.items():
            self._getFsFiles().update_many(
                {"_id": {"$in": [ObjectId(p) for p in params]}}, {"$pull": {"metadata.jobs": str(job)}}
            )

        # delete params
        all_params = [ObjectId(j) for i in file_params.values() for j in i]
        # for all of the params that can be deleted (with no jobs, and no lock), remove the sha256
        # In this way the file cannot be utilized as a cached file again
        self._getFsFiles().update_many(
            {"_id": {"$in": all_params}, "metadata.jobs": [], "metadata.tmp_lock": 0},
            {"$set": {"metadata.sha256": None}},
        )
        # now get the list of files to be deleted
        params_to_delete = self._getFsFiles().find(
            {"_id": {"$in": all_params}, "metadata.jobs": [], "metadata.tmp_lock": 0, "metadata.sha256": None}
        )
        # delete them
        for p in params_to_delete:
            self._getFs().delete(ObjectId(p["_id"]))

        # delete jobs
        self._getCollection().delete_many(job_query)


    def terminate_all_jobs(self):
        pass

class Job(object):
    def __init__(self, queue, data):
        """ """
        self._queue = queue
        self._data = data

    def __str__(self) -> str:
        return f"ID: {self.job_id}, created: {self.created_at}, finished: {self.finished_at}, result: {self.result}"

    @property
    def method(self):
        return self._data["payload"]["method"]

    @property
    def payload(self):
        return self._data["payload"]

    @property
    def job_id(self):
        return self._data["_id"]

    @property
    def number(self):
        return self._data["number"] if "number" in self._data else -1

    @property
    def priority(self):
        return self._data["priority"]

    @property
    def attempts_left(self):
        return self._data["attempts_left"]

    @property
    def is_failed(self):
        return self._data["attempts_left"] == 0

    @property
    def locked_by(self):
        return self._data["locked_by"]

    @property
    def locked_at(self):
        return self._data["locked_at"]

    @property
    def created_at(self):
        return self._data["created_at"]

    @property
    def started_at(self):
        return self._data["started_at"]

    @property
    def last_error(self):
        return self._data["last_error"]

    @property
    def finished_at(self):
        return self._data["finished_at"]

    @property
    def is_finished(self):
        return self._data["finished_at"] is not None

    @property
    def parameters(self):
        method_str = ""
        if "payload" in self._data and "params" in self._data["payload"] and "method" in self._data["payload"]:
            payload_params = json.loads(self._data["payload"]["params"])
            method_str = self._data["payload"]["method"]
            indexed_key_values = []
            named_key_values = []
            for k, v in payload_params.items():
                try:
                    int(k)
                    indexed_key_values.append(v)
                except:
                    named_key_values.append(v)
            combined_values = indexed_key_values + named_key_values
            method_str += "(" + ", ".join([str(v) for v in combined_values]) + ")"
        return method_str

    @property
    def progress(self):
        return self._data["progress"]


    # This is a GridFS id, not the actual result
    @property
    def result(self):
        return self._data["result"]

    @result.setter
    def result(self, res):
        self._data["result"] = res

    ## job control

    def complete(self, result=None):
        """job has been completed."""
        if result:
            self._data["result"] = result
        job = self._queue.collection.find_one_and_update(
            filter={"_id": self.job_id, "locked_by": self._queue.consumer_id},
            update={"$set": {"finished_at": datetime.now(), "result": self._data["result"], "progress": 1}},
            return_document=ReturnDocument.AFTER
        )
        self._queue._notify_dependent_jobs(str(job["_id"]))
        return job 

    def error(self, message=None):
        """note an error processing a job, and return it to the queue."""
        job = self._queue.collection.find_one_and_update(
            filter={"_id": self.job_id, "locked_by": self._queue.consumer_id},
            update={"$set": {"locked_by": None, "locked_at": None, "last_error": message}, "$inc": {"attempts_left": -1}},
            return_document=ReturnDocument.AFTER
        )
        if job["attempts_left"] <= 0:
            self._queue._notify_dependent_jobs(self, str(job["_id"]))

    def progressor(self, count=0):
        """note progress on a long running task."""
        return self._queue.collection.find_one_and_update(
            filter={"_id": self.job_id, "locked_by": self._queue.consumer_id},
            update={"$set": {"progress": count, "locked_at": datetime.now()}},
            return_document=ReturnDocument.AFTER
        )

    def release(self):
        """put the job back into_queue."""
        return self._queue.collection.find_one_and_update(
            filter={"_id": self.job_id, "locked_by": self._queue.consumer_id},
            update={"$set": {"locked_by": None, "locked_at": None}, "$inc": {"attempts_left": -1}},
            return_document=ReturnDocument.AFTER
        )

    ## context manager support

    def __enter__(self):
        return self._data

    def __exit__(self, type, value, tb):
        if (type, value, tb) == (None, None, None):
            self.complete()
        else:
            error = traceback.format_exc()
            self.error(error)

    # only works for jobs that report progress
    def _is_terminated(self, use_cached=False):
        if use_cached:
            data = self._data
        else:
            data = self._queue.collection.find_one({"_id": self.job_id})

        if data is None:
            return True

        return ("terminated" in data) and data["terminated"] or False

    # only works for jobs that report progress
    def terminate(self):
        return self._queue.collection.find_one_and_update(
            filter={"_id": self.job_id}, 
            update={"$set": {"terminated": True, "locked_at": datetime.now()}},
            return_document=ReturnDocument.AFTER
        )



#!/usr/bin/env python3

import logging
import random
from collections import Counter, defaultdict
from typing import List, Tuple, Dict

from mcrit.libs.utility import generate_segmented_sequence
from mcrit.minhash.MinHash import MinHash
from mcrit.minhash.ShingleLoader import ShingleLoader

logging.basicConfig(level=logging.INFO)
LOGGER = logging.getLogger(__name__)


class MinHasher(object):

    MINHASH_STRATEGY_HASH_ALL = 1
    MINHASH_STRATEGY_XOR_ALL = 2
    MINHASH_STRATEGY_SEGMENTED = 3

    def __init__(self, minhash_config, shingler_config):
        self._minhash_config = minhash_config
        self._shingler_config = shingler_config
        self._minhash_seeds = []
        self._minhash_permutation = []
        self._signature_segments = []
        self._shinglers = self._loadShinglers()
        self._shingler_names = [shingler.getName() for shingler in self._shinglers]
        self._initMinhashing()

    def _initMinhashing(self):
        random.seed(self._minhash_config.MINHASH_SEED)
        # initiate sequence of seeds. If XOR or SEGMENTED strategy, this serves as sequence of XOR values instead
        self._minhash_seeds = [
            random.randint(0, MinHash.getHashMax()) for _ in range(self._minhash_config.MINHASH_SIGNATURE_LENGTH)
        ]
        if self._minhash_config.MINHASH_STRATEGY == MinHasher.MINHASH_STRATEGY_SEGMENTED:
            self._initSegmentedMinHashing()

    def _initSegmentedMinHashing(self):
        # for weighing by segments, init segment fields
        weights = {shingler.getName(): shingler.getWeight() for shingler in self._shinglers}
        self._signature_segments = generate_segmented_sequence(weights, self._minhash_config.MINHASH_SIGNATURE_LENGTH)

    def _loadShinglers(self):
        shingle_loader = ShingleLoader(self._shingler_config)
        return shingle_loader.getShinglers()

    def isMinHashableFunction(self, smda_function):
        is_hashable = False
        if (
            self._minhash_config.MINHASH_FN_MIN_BLOCKS
            and smda_function.num_blocks > self._minhash_config.MINHASH_FN_MIN_BLOCKS
        ):
            is_hashable = True
        if (
            self._minhash_config.MINHASH_FN_MIN_INS
            and smda_function.num_instructions > self._minhash_config.MINHASH_FN_MIN_INS
        ):
            is_hashable = True
        return is_hashable

    def _calculateMinHash(self, smda_function):
        if self._minhash_config.MINHASH_STRATEGY == MinHasher.MINHASH_STRATEGY_HASH_ALL:
            return self._calculateMinHashAllSeeds(smda_function)
        elif self._minhash_config.MINHASH_STRATEGY == MinHasher.MINHASH_STRATEGY_XOR_ALL:
            return self._calculateMinHashAllXored(smda_function)
        elif self._minhash_config.MINHASH_STRATEGY == MinHasher.MINHASH_STRATEGY_SEGMENTED:
            return self._calculateMinHashSegmented(smda_function)
        raise NotImplementedError("Unknown MinHasher strategy.")

    def calculateMinHashFromStorage(self, function_tuple):
        function_id, smda_function = function_tuple
        minhash = self._calculateMinHash(smda_function)
        minhash.function_id = function_id
        return minhash

    def calculateMinHashesFromStorage(self, function_tuples):
        minhashes = []
        for function_tuple in function_tuples:
            function_id, smda_function = function_tuple
            minhash = self._calculateMinHash(smda_function)
            minhash.function_id = function_id
            minhashes.append(minhash)
        return minhashes

    def calculateAggregatedScoresFromPackedTuples(
        self,
        packed_tuples: List[Tuple[int, int, bytes, int, int, bytes]],
        ignore_threshold=False,
        minhash_threshold=None,
    ) -> Dict[Tuple[int, int, int], Tuple[int, float]]:
        results: Dict[Tuple[int, int, int], Tuple[int, float]] = {}
        if minhash_threshold is None:
            minhash_threshold = self._minhash_config.MINHASH_MATCHING_THRESHOLD
        for minhash_tuple in packed_tuples:
            sample_id_a, function_id_a, minhash_a, sample_id_b, function_id_b, minhash_b = minhash_tuple
            score = MinHash.calculateMinHashScore(
                minhash_a, minhash_b, minhash_bits=self._minhash_config.MINHASH_SIGNATURE_BITS
            )
            if ignore_threshold or score > minhash_threshold:
                key = (sample_id_a, function_id_a, sample_id_b)
                if key not in results or score > results[key][1]:
                    results[key] = (function_id_b, score)
        return results

    def calculateScoresFromPackedTuples(
        self,
        packed_tuples: List[Tuple[int, int, bytes, int, int, bytes]],
        ignore_threshold=False,
        minhash_threshold=None,
    ) -> List[Tuple[int, int, int, int, float]]:
        results = []
        if minhash_threshold is None:
            minhash_threshold = self._minhash_config.MINHASH_MATCHING_THRESHOLD
        for minhash_tuple in packed_tuples:
            sample_id_a, function_id_a, minhash_a, sample_id_b, function_id_b, minhash_b = minhash_tuple
            score = MinHash.calculateMinHashScore(
                minhash_a, minhash_b, minhash_bits=self._minhash_config.MINHASH_SIGNATURE_BITS
            )
            if ignore_threshold or score > minhash_threshold:
                results.append((sample_id_a, function_id_a, sample_id_b, function_id_b, score))
        return results

    def _calculateMinHashAllSeeds(self, smda_function):
        """Calculate hash function every time, then take minimum shingle per shingler"""
        minhash_result = MinHash(minhash_bits=self._minhash_config.MINHASH_SIGNATURE_BITS)
        minhash_signature = []
        shingler_composition = Counter()
        for hash_seed in self._minhash_config.HASH_SEEDS:
            minimum_shingles = []
            shingler_outputs = {}
            for shingler in self._shinglers:
                shingler_outputs[shingler.getName()] = shingler.process(smda_function, hash_seed)
            for shingler, outputs in shingler_outputs.items():
                merged_outputs = []
                for output in outputs:
                    merged_outputs.extend(output)
                minimum_shingle = merged_outputs[0]
                if len(merged_outputs) > 1:
                    minimum_shingle = min(merged_outputs)
                minimum_shingles.append(minimum_shingle)
            if self._minhash_config.MINHASH_TRACK_SHINGLES:
                shingler_composition[self._shingler_names[minimum_shingles.index(min(minimum_shingles))]] += 1
            minhash_value = min(minimum_shingles)
            if self._minhash_config.MINHASH_SIGNATURE_BITS < 32:
                minhash_value = minhash_value % (2 ** self._minhash_config.MINHASH_SIGNATURE_BITS)
            minhash_signature.append(minhash_value)
        minhash_result.setMinHash(minhash_signature)
        minhash_result.shingler_composition = dict(shingler_composition)
        return minhash_result

    def _calculateMinHashAllXored(self, smda_function):
        """Calculate group of shingles per shingler, then XOR all shingles and pick minimum across all combined.
        For multi byte-sequence shinglers, this will always pick the respective minimum shingle"""
        minhash_result = MinHash(minhash_bits=self._minhash_config.MINHASH_SIGNATURE_BITS)
        minhash_signature = []
        shingler_composition = {name: {"count": 0, "size": 0} for name in self._shingler_names}
        shingler_outputs = {}
        for shingler in self._shinglers:
            shingler_outputs[shingler.getName()] = shingler.process(smda_function, 0)
            shingler_composition[shingler.getName()]["size"] = sum(
                [len(item) for item in shingler_outputs[shingler.getName()]]
            )
        for hash_seed in self._minhash_seeds:
            xored_shingles = []
            for shingler, outputs in sorted(shingler_outputs.items()):
                xored_outputs = []
                for output in outputs:
                    xored_outputs.extend([shingle ^ hash_seed for shingle in output])
                minimum_shingle = xored_outputs[0]
                if len(xored_outputs) > 1:
                    minimum_shingle = min(xored_outputs)
                xored_shingles.append(minimum_shingle)
            if self._minhash_config.MINHASH_TRACK_SHINGLES:
                shingler_composition[self._shingler_names[xored_shingles.index(min(xored_shingles))]]["count"] += 1
            minhash_value = min(xored_shingles)
            if self._minhash_config.MINHASH_SIGNATURE_BITS < 32:
                minhash_value = minhash_value % (2 ** self._minhash_config.MINHASH_SIGNATURE_BITS)
            minhash_signature.append(minhash_value)
        minhash_result.setMinHash(minhash_signature)
        minhash_result.shingler_composition = dict(shingler_composition)
        return minhash_result

    def _calculateMinHashSegmented(self, smda_function):
        """Split the whole signature into segments according to weights and have shingles of the shingler
        only compete within the segment. Use XOR on all shingles to speed up the procedure"""
        minhash_result = MinHash(minhash_bits=self._minhash_config.MINHASH_SIGNATURE_BITS)
        minhash_signature = []
        shingler_composition = {}
        shingler_outputs = {}
        for shingler in self._shinglers:
            shingler.setUseWeights(False)
            shingler_name = shingler.getName()
            shingler_output = shingler.process(smda_function, 0)
            shingler_outputs[shingler_name] = shingler_output[0]
            shingler_size = len(shingler_output[0])
            if self._minhash_config.MINHASH_TRACK_SHINGLES:
                shingler_count = self._signature_segments.count(shingler_name)
                shingler_composition[shingler_name] = {"size": shingler_size, "count": shingler_count}
        for index, shingler_name in enumerate(self._signature_segments):
            hash_seed = self._minhash_seeds[index]
            xored_outputs = [shingle ^ hash_seed for shingle in shingler_outputs[shingler_name]]
            minhash_value = min(xored_outputs)
            if self._minhash_config.MINHASH_SIGNATURE_BITS < 32:
                minhash_value = minhash_value % (2 ** self._minhash_config.MINHASH_SIGNATURE_BITS)
            minhash_signature.append(minhash_value)
        minhash_result.setMinHash(minhash_signature)
        minhash_result.shingler_composition = dict(shingler_composition)
        return minhash_result


import math
from copy import deepcopy
from typing import TYPE_CHECKING, Dict, List, Optional

from smda.common.SmdaFunction import SmdaFunction
from smda.common.BinaryInfo import BinaryInfo

from mcrit.libs.graph import Graph
from mcrit.storage.SampleEntry import SampleEntry
from mcrit.storage.MatchedSampleEntry import MatchedSampleEntry
from mcrit.storage.MatchedFunctionEntry import MatchedFunctionEntry
import mcrit.matchers.MatcherInterface as MatcherInterface

if TYPE_CHECKING:  # pragma: no cover
    from mcrit.storage.SampleEntry import SampleEntry

# Dataclass, post init
# constructor -> .fromSmdaFunction
# assume sample_entry, smda_function always available

class MatchingResult(object):
    reference_sample_entry: "SampleEntry"
    other_sample_entry: "SampleEntry"
    match_aggregation: Dict
    sample_matches: List["MatchedSampleEntry"]
    function_matches: List["MatchedFunctionEntry"]
    # filtered versions 
    filtered_sample_matches: List["MatchedSampleEntry"]
    filtered_function_matches: List["MatchedFunctionEntry"]
    # function_id -> [(family_id, sample_id), ...]
    library_matches: Dict
    # function_id -> {family_id_a, family_id_b, ...}
    function_id_to_family_ids_matched: Dict
    unique_family_scores_per_sample: Dict
    family_id_to_name_map: Dict
    is_family_filtered: bool
    is_sample_filtered: bool
    is_function_filtered: bool
    is_score_filtered: bool
    is_sample_count_filtered: bool
    is_family_count_filtered: bool
    is_library_filtered: bool
    is_pic_filtered: bool
    is_query: bool
    filter_values: Dict

    def __init__(self, sample_entry: "SampleEntry") -> None:
        self.reference_sample_entry = sample_entry
        self.unique_family_scores_per_sample = None
        self.family_id_to_name_map = None
        self.is_query = False
        self.filter_values = {}
        self.function_id_to_family_ids_matched = {}

    @property
    def num_original_function_matches(self):
        return len(self.function_matches)
    @property
    def num_original_sample_matches(self):
        return len(self.sample_matches)
    @property
    def num_original_family_matches(self):
        return len(set([sample.family_id for sample in self.sample_matches if not sample.is_library]))
    @property
    def num_original_library_matches(self):
        return len(set([sample.family_id for sample in self.sample_matches if sample.is_library]))
    @property
    def num_function_matches(self):
        return len(self.filtered_function_matches)
    @property
    def num_sample_matches(self):
        return len(self.filtered_sample_matches)
    @property
    def num_family_matches(self):
        return len(set([sample.family_id for sample in self.filtered_sample_matches if not sample.is_library]))
    @property
    def num_library_matches(self):
        return len(set([sample.family_id for sample in self.filtered_sample_matches if sample.is_library]))

    def setFilterValues(self, filter_dict):
        self.filter_values = filter_dict

    def getFilterValue(self, filter_name):
        return self.filter_values[filter_name] if filter_name in self.filter_values else 0
    
    def applyFilterValues(self):
        """ use the filter_values that have been set before to reduce the data """
        # filter family/sample
        if self.filter_values.get("filter_direct_min_score", None):
            self.filterToDirectMinScore(self.filter_values["filter_direct_min_score"])
        if self.filter_values.get("filter_direct_nonlib_min_score", None):
            self.filterToDirectMinScore(self.filter_values["filter_direct_nonlib_min_score"], nonlib=True)
        if self.filter_values.get("filter_frequency_min_score", None):
            self.filterToFrequencyMinScore(self.filter_values["filter_frequency_min_score"])
        if self.filter_values.get("filter_frequency_nonlib_min_score", None):
            self.filterToFrequencyMinScore(self.filter_values["filter_frequency_nonlib_min_score"], nonlib=True)
        if self.filter_values.get("filter_unique_only", None):
            self.filterToUniqueMatchesOnly()
        if self.filter_values.get("filter_exclude_own_family", None):
            self.excludeOwnFamily()
        if self.filter_values.get("filter_family_name", None):
            self.filterByFamilyName(self.filter_values["filter_family_name"])
        # filter functions
        if self.filter_values.get("filter_exclude_library", None):
            self.excludeLibraryMatches()
        if self.filter_values.get("filter_library_min_score", None):
            self.filterToFunctionScore(min_score=self.filter_values["filter_library_min_score"], library_only=True)
        if self.filter_values.get("filter_max_num_families", None):
            self.filterToFamilyCount(self.filter_values["filter_max_num_families"])
        if self.filter_values.get("filter_max_num_samples", None):
            self.filterToSampleCount(self.filter_values["filter_max_num_samples"])
        if self.filter_values.get("filter_function_min_score", None):
            self.filterToFunctionScore(min_score=self.filter_values["filter_function_min_score"])
        if self.filter_values.get("filter_function_max_score", None):
            self.filterToFunctionScore(max_score=self.filter_values["filter_function_max_score"])
        if self.filter_values.get("filter_function_offset", None):
            self.filterToFunctionOffset(self.filter_values["filter_function_offset"])
        if self.filter_values.get("filter_exclude_pic", None):
            self.excludePicMatches()
        if self.filter_values.get("filter_func_unique", None):
            self.filterToUniqueFunctionMatchesOnly()

    def getFamilyNameByFamilyId(self, family_id):
        if self.family_id_to_name_map is None:
            self.family_id_to_name_map = {}
            for sample_match in self.sample_matches:
                self.family_id_to_name_map[sample_match.family_id] = sample_match.family
        return self.family_id_to_name_map[family_id] if family_id in self.family_id_to_name_map else ""
    
    def getFamilyIdsMatchedByFunctionId(self, function_id):
        if function_id not in self.function_id_to_family_ids_matched:
            return 0
        return self.function_id_to_family_ids_matched[function_id]

    def filterByFamilyName(self, filter_term):
        """ reduce families and samples to those where family_name is part of the family_name """
        filtered_sample_matches = []
        for sample_match in self.filtered_sample_matches:
            if filter_term in sample_match.family:
                filtered_sample_matches.append(sample_match)
        self.filtered_sample_matches = filtered_sample_matches

    def filterToDirectMinScore(self, min_score, nonlib=False):
        """ reduce aggregated sample matches to those with direct score of min_score or higher, but nonlib flag is not applied to library samples """
        filtered_sample_matches = []
        for sample_match in self.filtered_sample_matches:
            if nonlib:
                if sample_match.is_library:
                    filtered_sample_matches.append(sample_match)
                elif sample_match.matched_percent_nonlib_score_weighted >= min_score:
                    filtered_sample_matches.append(sample_match)
            else:
                if sample_match.matched_percent_score_weighted >= min_score:
                    filtered_sample_matches.append(sample_match)
        self.filtered_sample_matches = filtered_sample_matches

    def filterToFrequencyMinScore(self, min_score, nonlib=False):
        """ reduce aggregated sample matches to those with frequency score of min_score or higher, but nonlib flag is not applied to library samples """
        filtered_sample_matches = []
        for sample_match in self.filtered_sample_matches:
            if nonlib:
                if sample_match.is_library:
                    filtered_sample_matches.append(sample_match)
                elif sample_match.matched_percent_nonlib_frequency_weighted >= min_score:
                    filtered_sample_matches.append(sample_match)
            else:
                if sample_match.matched_percent_frequency_weighted >= min_score:
                    filtered_sample_matches.append(sample_match)
        self.filtered_sample_matches = filtered_sample_matches

    def filterToUniqueMatchesOnly(self):
        """ reduce aggregated sample matches to those with unique matches only """
        filtered_sample_matches = []
        for sample_match in self.filtered_sample_matches:
            unique_info = self.getUniqueFamilyMatchInfoForSample(sample_match.sample_id)
            if unique_info["unique_score"] > 0 or sample_match.is_library:
                filtered_sample_matches.append(sample_match)
        self.filtered_sample_matches = filtered_sample_matches

    def filterToUniqueFunctionMatchesOnly(self):
        """ reduce function matches to those with unique matches (with respect to the family) only """
        aggregated = self.getAggregatedFunctionMatches()
        filtered_function_matches = []
        unique_info_by_function_id = {entry["function_id"]: entry["num_families_matched"] == 1 for entry in aggregated}
        for function_match in self.filtered_function_matches:
            if unique_info_by_function_id[function_match.function_id]:
                filtered_function_matches.append(function_match)
        self.filtered_function_matches = filtered_function_matches

    def filterToFunctionOffset(self, offset):
        """ reduce function matches to those that match a specific offset """
        filtered_function_matches = []
        for function_match in self.filtered_function_matches:
            if function_match.offset == offset:
                filtered_function_matches.append(function_match)
        self.filtered_function_matches = filtered_function_matches

    def excludeOwnFamily(self):
        """ remove all sample matches with the same family_id as the reference samples"""
        filtered_sample_matches = []
        for sample_match in self.filtered_sample_matches:
            if sample_match.family_id != self.reference_sample_entry.family_id:
                filtered_sample_matches.append(sample_match)
        self.filtered_sample_matches = filtered_sample_matches

    def excludeLibraryMatches(self):
        """ reduce contained matches to those where none of the matches is with a library (transitive library identification) """
        library_matches = [matches for matches in self.library_matches.values() if matches]
        library_samples = set([match[1] for match_list in library_matches for match in match_list])
        library_matched_functions = [key for key in self.library_matches if self.library_matches[key]]
        self.filtered_sample_matches = [sample_match for sample_match in self.filtered_sample_matches if sample_match.sample_id not in library_samples]
        self.filtered_function_matches = [function_match for function_match in self.filtered_function_matches if function_match.function_id not in library_matched_functions]
        self.is_library_filtered = True

    def excludePicMatches(self):
        """ reduce contained matches to those which are not identified as quasi-identical via PIC matching """
        self.filtered_function_matches = [function_match for function_match in self.filtered_function_matches if not function_match.match_is_pichash]
        self.is_pic_filtered = True

    def filterToSampleCount(self, max_sample_count):
        """ reduce contained matches to those with a maximum of <max_sample_count> matched samples """
        matched_samples_by_function_id = {}
        for function_match in self.filtered_function_matches:
            if not function_match.function_id in matched_samples_by_function_id:
                matched_samples_by_function_id[function_match.function_id] = []
            if not function_match.matched_sample_id in matched_samples_by_function_id[function_match.function_id]:
                matched_samples_by_function_id[function_match.function_id].append(function_match.matched_family_id)
        self.filtered_function_matches = [function_match for function_match in self.filtered_function_matches if len(matched_samples_by_function_id[function_match.function_id]) <= max_sample_count]
        self.is_sample_count_filtered = True

    def filterToFamilyCount(self, max_family_count):
        """ reduce contained matches to those with a maximum of <max_family_count> matched families """
        matched_families_by_function_id = {}
        for function_match in self.filtered_function_matches:
            if not function_match.function_id in matched_families_by_function_id:
                matched_families_by_function_id[function_match.function_id] = []
            if not function_match.matched_family_id in matched_families_by_function_id[function_match.function_id]:
                matched_families_by_function_id[function_match.function_id].append(function_match.matched_family_id)
        self.filtered_function_matches = [function_match for function_match in self.filtered_function_matches if len(matched_families_by_function_id[function_match.function_id]) <= max_family_count]
        self.is_family_count_filtered = True

    def filterToFunctionScore(self, min_score=None, max_score=None, library_only=False):
        """ reduce contained matches to those with a minimum score of <threshold> """
        filtered_function_matches = []
        for function_match in self.filtered_function_matches:
            if library_only and not function_match.match_is_library:
                filtered_function_matches.append(function_match)
            else:
                if (min_score is not None) and (max_score is not None):
                    if min_score <= function_match.matched_score <= max_score:
                        filtered_function_matches.append(function_match)
                else:
                    # no duplicates expected as per above condition only one case can be true
                    if min_score is not None and function_match.matched_score >= min_score:
                        filtered_function_matches.append(function_match)
                    if max_score is not None and function_match.matched_score <= max_score:
                        filtered_function_matches.append(function_match)
        self.filtered_function_matches = filtered_function_matches
        self.is_score_filtered = True

    def filterToFamilyId(self, family_id):
        """ reduce contained matches to chosen family_id by deleting the other sample and function matches """
        self.filtered_sample_matches = [sample_match for sample_match in self.filtered_sample_matches if sample_match.family_id == family_id]
        self.filtered_function_matches = [function_match for function_match in self.filtered_function_matches if function_match.matched_family_id == family_id]
        self.is_family_filtered = True

    def filterToSampleId(self, sample_id):
        """ reduce contained matches to chosen sample_id by deleting the other sample and function matches """
        self.filtered_sample_matches = [sample_match for sample_match in self.filtered_sample_matches if sample_match.sample_id == sample_id]
        self.filtered_function_matches = [function_match for function_match in self.filtered_function_matches if function_match.matched_sample_id == sample_id]
        self.is_sample_filtered = True

    def filterToFunctionId(self, function_id):
        """ reduce contained matches to chosen function_id by deleting the other sample and function matches """
        self.filtered_function_matches = [function_match for function_match in self.filtered_function_matches if function_match.function_id == function_id]
        self.is_function_filtered = True

    def hasLibraryMatch(self, function_id):
        return function_id in self.library_matches and self.library_matches[function_id]

    def getBestSampleMatchesPerFamily(self, start=None, limit=None, unfiltered=False, library_only=False, malware_only=False):
        by_family = {}
        source_matches = self.sample_matches if unfiltered else self.filtered_sample_matches
        for sample_match in source_matches:
            if library_only and not sample_match.is_library:
                continue
            if malware_only and sample_match.is_library:
                continue
            if sample_match.family not in by_family:
                by_family[sample_match.family] = {
                    "score": sample_match.matched_percent_frequency_weighted,
                    "report": sample_match
                }
            elif sample_match.matched_percent_frequency_weighted > by_family[sample_match.family]["score"]:
                by_family[sample_match.family]["score"] = sample_match.matched_percent_frequency_weighted
                by_family[sample_match.family]["report"] = sample_match
        result_list = []
        for family, score_entry in sorted(by_family.items(), key=lambda e: e[1]["score"], reverse=True):
            result_list.append(score_entry["report"])
        if start is not None:
            result_list = result_list[start:]
        if limit is not None:
            result_list = result_list[:limit]
        return result_list

    def getUniqueFamilyMatchInfoForSample(self, sample_id):
        if self.unique_family_scores_per_sample is None:
            self.unique_family_scores_per_sample = {entry.sample_id: {"functions_matched": 0, "bytes_matched": 0, "unique_score": 0} for entry in self.sample_matches}
            families_matched_by_function_id = {}
            samples_matched_by_function_id = {}
            weighted_bytes_per_function_id = {}
            for function_match_summary in self.function_matches:
                if function_match_summary.function_id not in families_matched_by_function_id:
                    families_matched_by_function_id[function_match_summary.function_id] = set()
                    samples_matched_by_function_id[function_match_summary.function_id] = set()
                    weighted_bytes_per_function_id[function_match_summary.function_id] = 0
                families_matched_by_function_id[function_match_summary.function_id].add(function_match_summary.matched_family_id)
                samples_matched_by_function_id[function_match_summary.function_id].add(function_match_summary.matched_sample_id)
                # matches should be weighted by match score
                weighted_bytes_per_function_id[function_match_summary.function_id] = function_match_summary.num_bytes * function_match_summary.matched_score / 100.0
            for function_id in families_matched_by_function_id:
                if len(families_matched_by_function_id[function_id]) == 1:
                    for sid in samples_matched_by_function_id[function_id]:
                        self.unique_family_scores_per_sample[sid]["functions_matched"] += 1
                        self.unique_family_scores_per_sample[sid]["bytes_matched"] += weighted_bytes_per_function_id[function_id]
            for sid in self.unique_family_scores_per_sample:
                self.unique_family_scores_per_sample[sid]["unique_score"] = 100.0 * self.unique_family_scores_per_sample[sid]["bytes_matched"] / self.reference_sample_entry.binweight
        if sample_id in self.unique_family_scores_per_sample:
            return self.unique_family_scores_per_sample[sample_id]
        else:
            return {"functions_matched": 0, "bytes_matched": 0, "unique_score": 0}

    def getSampleMatches(self, start=None, limit=None, unfiltered=False, library_only=False, malware_only=False):
        by_sample_id = {}
        source_matches = self.sample_matches if unfiltered else self.filtered_sample_matches
        for sample_match in source_matches:
            if library_only and not sample_match.is_library:
                continue
            if malware_only and sample_match.is_library:
                continue
            by_sample_id[sample_match.sample_id] = {
                "score": sample_match.matched_percent_frequency_weighted,
                "report": sample_match
            }
        result_list = []
        for sample_id, score_entry in sorted(by_sample_id.items(), key=lambda e: e[1]["score"], reverse=True):
            result_list.append(score_entry["report"])
        if start is not None:
            result_list = result_list[start:]
        if limit is not None:
            result_list = result_list[:limit]
        return result_list
    
    def getFunctionMatches(self, start=None, limit=None, unfiltered=False):
        source_matches = self.function_matches if unfiltered else self.filtered_function_matches
        result_list = source_matches
        if start is not None:
            result_list = result_list[start:]
        if limit is not None:
            result_list = result_list[:limit]
        return result_list

    def getAggregatedFunctionMatches(self, start=None, limit=None, unfiltered=False):
        by_function_id = {}
        source_matches = self.function_matches if unfiltered else self.filtered_function_matches
        for function_match in source_matches:
            if function_match.function_id not in by_function_id:
                by_function_id[function_match.function_id] = {
                    "function_id": function_match.function_id,
                    "num_bytes": 0,
                    "offset": 0,
                    "best_score": 0,
                    "num_families_matched": 0,
                    "family_ids_matched": set([]),
                    "families_matched": set([]),
                    "num_samples_matched": 0,
                    "sample_ids_matched": set([]),
                    "num_functions_matched": 0,
                    "function_ids_matched": set([]),
                    "minhash_matches": 0,
                    "pichash_matches": 0,
                    "library_matches": 0,
                }
            by_function_id[function_match.function_id]["num_bytes"] = function_match.num_bytes
            by_function_id[function_match.function_id]["offset"] = function_match.offset
            by_function_id[function_match.function_id]["best_score"] = max(function_match.matched_score, by_function_id[function_match.function_id]["best_score"])
            by_function_id[function_match.function_id]["family_ids_matched"].add(function_match.matched_family_id)
            by_function_id[function_match.function_id]["families_matched"].add(self.getFamilyNameByFamilyId(function_match.matched_family_id))
            by_function_id[function_match.function_id]["sample_ids_matched"].add(function_match.matched_sample_id)
            by_function_id[function_match.function_id]["function_ids_matched"].add(function_match.matched_function_id)
            by_function_id[function_match.function_id]["num_families_matched"] = len(by_function_id[function_match.function_id]["family_ids_matched"])
            by_function_id[function_match.function_id]["num_samples_matched"] = len(by_function_id[function_match.function_id]["sample_ids_matched"])
            by_function_id[function_match.function_id]["num_functions_matched"] = len(by_function_id[function_match.function_id]["function_ids_matched"])
            by_function_id[function_match.function_id]["minhash_matches"] += 1 if function_match.match_is_minhash else 0
            by_function_id[function_match.function_id]["pichash_matches"] += 1 if function_match.match_is_pichash else 0
            by_function_id[function_match.function_id]["library_matches"] += 1 if function_match.match_is_library else 0
            by_function_id[function_match.function_id]["is_family_unique"] = True if len(by_function_id[function_match.function_id]["family_ids_matched"]) == 1 else False
        aggregated_matched = [v for k, v in sorted(by_function_id.items())]
        if start is not None:
            aggregated_matched = aggregated_matched[start:]
        if limit is not None:
            aggregated_matched = aggregated_matched[:limit]
        return aggregated_matched

    def getFunctionsSlice(self, start, limit, unfiltered=False):
        return self.filtered_function_matches[start:start+limit]
    
    def getLinkHuntResults(self, min_score=None, min_lib_score=None, min_size=None, min_offset=None, max_offset=None, unpenalized_family_count=3, exclude_family_ids=None, exclude_sample_ids=None, strongest_per_family=True):
        """ Returns the most promising matches for finding inter-family relationship as a list """
        # run over all function_matches while applying filters, collect aggregate data needed to calculate scores
        by_function_id = {}
        link_matches = []
        droppable_fids = set()
        for function_match in sorted(self.function_matches, key=lambda x: (x.matched_score, x.matched_family_id), reverse=True):
            if min_offset is not None and function_match.offset < min_offset:
                continue
            if max_offset is not None and max_offset < function_match.offset:
                continue
            if min_lib_score is not None and function_match.match_is_library and function_match.matched_score < min_lib_score:
                continue
            if min_score is not None and function_match.matched_score < min_score:
                continue
            if min_size is not None and function_match.num_bytes < min_size:
                continue
            if exclude_family_ids is not None and function_match.matched_family_id in exclude_family_ids:
                continue
            if exclude_sample_ids is not None and function_match.matched_sample_id in exclude_sample_ids:
                continue
            if function_match.function_id not in by_function_id:
                by_function_id[function_match.function_id] = {
                    "function_id": function_match.function_id,
                    "num_bytes": 0,
                    "offset": 0,
                    "position": len(by_function_id),
                    "num_families_matched": 0,
                    "family_ids_matched": set([]),
                    "families_matched": set([]),
                    "library_matches": 0,
                }
            # only keep the first / strongest per function_id and family_id
            if strongest_per_family and function_match.matched_family_id in by_function_id[function_match.function_id]["family_ids_matched"]:
                continue
            by_function_id[function_match.function_id]["num_bytes"] = function_match.num_bytes
            by_function_id[function_match.function_id]["offset"] = function_match.offset
            by_function_id[function_match.function_id]["family_ids_matched"].add(function_match.matched_family_id)
            by_function_id[function_match.function_id]["num_families_matched"] = len(by_function_id[function_match.function_id]["family_ids_matched"])
            family_name = self.getFamilyNameByFamilyId(function_match.matched_family_id)
            function_match.matched_family = family_name
            by_function_id[function_match.function_id]["families_matched"].add(family_name)
            by_function_id[function_match.function_id]["library_matches"] += 1 if function_match.match_is_library else 0
            # these will be discarded later on anyway
            if function_match.match_is_library:
                droppable_fids.add(function_match.function_id)
            link_matches.append(function_match)
        filtered_link_matched = []
        for match in link_matches:
            if match.function_id in droppable_fids:
                continue
            filtered_link_matched.append(match)
        link_matches = filtered_link_matched
        for match in link_matches:
            num_families_matched = by_function_id[match.function_id]["num_families_matched"]
            family_adjustment_value = 1 if num_families_matched < unpenalized_family_count else 1 + int(math.log(1 + num_families_matched - unpenalized_family_count, 2))
            # for now, we use an interpolated polynome that is flatter than linear initially and crosses upwards once past ~10 families matched
            family_adjustment_value = 1 if num_families_matched < unpenalized_family_count else max(1, math.floor(0.0545793*num_families_matched**2 + 0.370159*num_families_matched - 0.170796))
            # magic formula: 
            # do a compound score that is based on
            # 10% position, front of binary preferred
            # 20% size, functions larger 100bytes preferred
            # 70% score, as originally produced by minhash
            position_score = 100 * (1 - (by_function_id[match.function_id]["position"] / len(by_function_id)))
            size_score = 100 * (min(300, by_function_id[match.function_id]["num_bytes"]) / 300)
            # scale the whole compound score with frequency weight
            match.matched_link_score = (0.1 * position_score + 0.2 * size_score + 0.7 * match.matched_score) / family_adjustment_value
            match.matched_unique = num_families_matched == 1
            link_matches.sort(key=lambda x: (x.matched_link_score, x.function_id), reverse=True)
        return link_matches
    
    def clusterLinkHuntResult(self, function_entries, link_hunt_matches):
        """ 
        Clustering is based on ICFG relationships, so we need to pass in that information in the form of function_entries
        """
        cluster_results = []
        # build some helper lookup structures
        function_offsets = [f.offset for f in function_entries]
        candidate_links = {}
        family_id_to_name = {}
        families_per_offset = {}
        for fm in link_hunt_matches:
            # map names
            family_id_to_name[fm.matched_family_id] = fm.matched_family
            # filter to some minimal link score (remove super common functions to break up huge clusters)
            if fm.matched_family_id not in candidate_links:
                candidate_links[fm.matched_family_id] = []
            if fm.matched_link_score > 1:
                candidate_links[fm.matched_family_id].append(fm)
            # make unique matches identifiable
            if fm.offset not in families_per_offset:
                families_per_offset[fm.offset] = set()
            families_per_offset[fm.offset].add(fm.matched_family_id)
        # extract code references from function_entries via SmdaFunctions
        binfo = BinaryInfo(b"")
        all_function_links = {}
        for function_entry in function_entries:
            smda_function = SmdaFunction.fromDict(function_entry.xcfg, binfo)
            for _, to_offsets in smda_function.outrefs.items():
                for to_offset in [o for o in to_offsets if o in function_offsets]:
                    from_offset = smda_function.offset
                    if from_offset not in all_function_links:
                        all_function_links[from_offset] = set()
                    all_function_links[from_offset].add(to_offset)
                    if to_offset not in all_function_links:
                        all_function_links[to_offset] = set()
                    all_function_links[to_offset].add(from_offset)
        # turn into proper (undirected) ICFG and DFS all clusters
        for fam_id, links in candidate_links.items():
            offset_to_link = {}
            g = Graph()
            for index, link_a in enumerate(links):
                g.addNode(link_a.offset)
                offset_to_link[link_a.offset] = link_a
                for link_b in links[index:]:
                    if link_a.offset in all_function_links and link_b.offset in all_function_links[link_a.offset]:
                        g.addEdge(link_a.offset, link_b.offset)
                        g.addEdge(link_b.offset, link_a.offset)
            queue = set([k for k in g.graph.keys()])
            while queue:
                start = queue.pop()
                reachable = g.DFS(start)
                subgraph_score = 0
                max_score = 0
                for offset in reachable:
                    subgraph_score += offset_to_link[offset].matched_link_score
                    max_score = max(max_score, offset_to_link[offset].matched_link_score)
                subgraph_result = {
                    "family_id": fam_id,
                    "family": family_id_to_name[fam_id],
                    "links": [offset_to_link[offset] for offset in reachable],
                    "unique_matches": [offset_to_link[offset] for offset in reachable if len(families_per_offset[offset]) == 1],
                    "score": subgraph_score,
                    "max_score": max_score
                }
                cluster_results.append(subgraph_result)
                queue.difference_update(reachable)
        return cluster_results

    def toDict(self):
        # we need to aggregate by function_id here
        summarized_function_match_summaries = {}
        for function_match_entry in self.function_matches:
            if function_match_entry.function_id not in summarized_function_match_summaries:
                summarized_function_match_summaries[function_match_entry.function_id] = {
                    "num_bytes": function_match_entry.num_bytes,
                    "offset": function_match_entry.offset,
                    "fid": function_match_entry.function_id,
                    "matches": [function_match_entry.getMatchTuple()]
                }
            else:
                summarized_function_match_summaries[function_match_entry.function_id]["matches"].append(function_match_entry.getMatchTuple())
        # build the dictionary
        matching_entry = {
            "info": {
                "sample": self.reference_sample_entry.toDict()
            },
            "matches": {
                "aggregation": self.match_aggregation,
                "functions": summarized_function_match_summaries,
                "samples": [match.toDict() for match in self.sample_matches]
            }
        }
        if self.other_sample_entry is not None:
            matching_entry["other_sample_info"] = self.other_sample_entry.toDict()
        return matching_entry

    @classmethod
    def fromDict(cls, entry_dict):
        matching_entry = cls(None)
        matching_entry.reference_sample_entry = SampleEntry.fromDict(entry_dict["info"]["sample"])
        if "other_sample_info" in entry_dict:
            matching_entry.other_sample_entry = SampleEntry.fromDict(entry_dict["other_sample_info"])
        else:
            matching_entry.other_sample_entry = None
        matching_entry.match_aggregation = entry_dict["matches"]["aggregation"]
        matching_entry.sample_matches = [MatchedSampleEntry.fromDict(entry) for entry in entry_dict["matches"]["samples"]]
        # expand function matches into individual entries
        list_of_function_matches = []
        matching_entry.library_matches = {abs(entry["fid"]): [] for entry in entry_dict["matches"]["functions"]}
        matching_entry.unique_family_scores_per_sample = None
        for function_match_summary in entry_dict["matches"]["functions"]:
            num_bytes = function_match_summary["num_bytes"]
            offset = function_match_summary["offset"]
            function_id = function_match_summary["fid"]
            # ensure that we have all function_ids in increasing order, regardless of whether they come from a query or regular match.
            matching_entry.is_query = True if function_id < 0 else False
            function_id = abs(function_id)
            if function_id not in matching_entry.function_id_to_family_ids_matched:
                matching_entry.function_id_to_family_ids_matched[function_id] = []
            for match_tuple in function_match_summary["matches"]:
                list_of_function_matches.append(MatchedFunctionEntry(function_id, num_bytes, offset, match_tuple))
                if match_tuple[0] not in matching_entry.function_id_to_family_ids_matched[function_id]:
                    matching_entry.function_id_to_family_ids_matched[function_id].append(match_tuple[0])
                if match_tuple[4] & MatcherInterface.IS_LIBRARY_FLAG:
                    if (match_tuple[0], match_tuple[1]) not in matching_entry.library_matches[function_id]:
                        matching_entry.library_matches[function_id].append((match_tuple[0], match_tuple[1]))
        matching_entry.function_matches = sorted(list_of_function_matches, key=lambda x: x.function_id)
        # create deep copies for filtering
        matching_entry.filtered_function_matches = deepcopy(matching_entry.function_matches)
        matching_entry.filtered_sample_matches = deepcopy(matching_entry.sample_matches)
        return matching_entry

    def __str__(self):
        return "Matched: Samples: {} (unfiltered: {}) Functions: {} (unfiltered: {})".format(
            len(self.filtered_sample_matches),
            len(self.sample_matches),
            len(self.filtered_function_matches),
            len(self.function_matches),
        )
